from collections import deque

from hwt.simulator.agentBase import SyncAgentBase
from hwtSimApi.process_utils import OnRisingCallbackLoop
from hwtSimApi.triggers import Timer, WaitWriteOnly, WaitCombRead, WaitCombStable,\
    WaitTimeslotEnd
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.constants import CLK_PERIOD


class FifoReaderAgent(SyncAgentBase):
    """
    Simulation agent for FifoReader interface
    """

    def __init__(self, sim, intf, allowNoReset=False):
        super(FifoReaderAgent, self).__init__(sim, intf, allowNoReset)
        self.data = deque()
        self.readPending = False
        self.lastData = None

        # flags to keep data coherent when enable state changes
        self.lastData_invalidate = False
        self.readPending_invalidate = False
        if intf.DATA_WIDTH == 0:
            raise NotImplementedError()

    def setEnable_asDriver(self, en):
        lastEn = self._enabled
        super(FifoReaderAgent, self).setEnable_asDriver(en)
        self.intf.wait.write(not en)
        self.lastData_invalidate = not en
        if not lastEn:
            self.dataWriter.setEnable(en)

    def setEnable_asMonitor(self, en):
        lastEn = self._enabled
        super(FifoReaderAgent, self).setEnable_asMonitor(en)
        self.intf.en.write(en)
        self.readPending_invalidate = not en
        if not lastEn:
            self.dataReader.setEnable(en)
        # else dataReader will disable itself

    def driver_init(self):
        yield WaitWriteOnly()
        self.intf.wait.write(not self._enabled)

    def monitor_init(self):
        yield WaitWriteOnly()
        self.intf.en.write(self._enabled)

    def get_data(self):
        return self.intf.data.read()

    def dataReader(self):
        yield Timer(1)
        if self.readPending:
            yield WaitCombRead()
            d = self.get_data()
            self.data.append(d)

            if self.readPending_invalidate:
                self.readPending = False
        if not self.readPending and not self._enabled:
            self.dataWriter.setEnable(False)

    def getMonitors(self):
        self.dataReader = OnRisingCallbackLoop(self.sim, self.clk,
                                               self.dataReader,
                                               self.getEnable)
        return ([self.monitor_init()] +
                super(FifoReaderAgent, self).getMonitors() +
                [self.dataReader()])

    def monitor(self):
        """
        Initialize data reading if wait is 0
        """
        intf = self.intf
        yield WaitCombRead()
        if self.notReset():
            # wait until wait signal is stable
            wait_last = None
            while True:
                yield WaitCombRead()
                wait = intf.wait.read()
                try:
                    wait = int(wait)
                except ValueError:
                    raise AssertionError(self.sim.now, intf, "wait signal in invalid state")

                if wait is wait_last:
                    break
                else:
                    wait_last = wait
                    yield WaitWriteOnly()

            rd = not wait
        else:
            rd = False

        yield WaitWriteOnly()
        intf.en.write(rd)

        self.readPending = rd

    def getDrivers(self):
        self.dataWriter = OnRisingCallbackLoop(self.sim, self.clk,
                                               self.dataWriter,
                                               self.getEnable)
        return ([self.driver_init()] +
                super(FifoReaderAgent, self).getDrivers() +
                [self.dataWriter()])

    def set_data(self, d):
        self.intf.data.write(d)

    def dataWriter(self):
        # delay data litle bit to have nicer wave
        # otherwise wirte happens before next clk period
        # and it means in 0 time and we will not be able to see it in wave
        yield Timer(1)
        yield WaitWriteOnly()
        self.set_data(self.lastData)
        if self.lastData_invalidate:
            self.lastData = None

        if not self._enabled:
            self.dataWriter.setEnable(False)

    def driver(self):
        # now we are before clock event
        # * set wait signal
        # * set last data (done in separate process)
        # * if en == 1, pop next data for next clk
        intf = self.intf
        yield WaitCombRead()
        rst_n = self.notReset()
        # speculative write
        if rst_n and self.data:
            wait = 0
        else:
            wait = 1

        yield WaitWriteOnly()
        intf.wait.write(wait)

        if rst_n:
            # wait for potential update of en
            # check if write can be performed and if it possible do real write
            yield WaitTimeslotEnd()
            en = intf.en.read()
            try:
                en = int(en)
            except ValueError:
                raise AssertionError(self.sim.now, intf,
                                     "en signal in invalid state")

            if en:
                assert self.data, (self.sim.now, intf, "underflow")
                self.lastData = self.data.popleft()


class FifoWriterAgent(SyncAgentBase):
    """
    Simulation agent for FifoWriter interface
    """

    def __init__(self, sim: HdlSimulator, intf, allowNoReset=False):
        super(FifoWriterAgent, self).__init__(
            sim, intf, allowNoReset=allowNoReset)
        self.data = deque()
        if intf.DATA_WIDTH == 0:
            raise NotImplementedError()

    def driver_init(self):
        yield WaitWriteOnly()
        self.intf.en.write(self._enabled)

    def monitor_init(self):
        yield WaitWriteOnly()
        self.intf.wait.write(not self._enabled)

    def setEnable_asDriver(self, en):
        SyncAgentBase.setEnable_asDriver(self, en)
        self.intf.en.write(en)

    def setEnable_asMonitor(self, en):
        SyncAgentBase.setEnable_asMonitor(self, en)
        self.intf.wait.write(not en)

    def get_data(self):
        return self.intf.data.read()

    def set_data(self, d):
        self.intf.data.write(d)

    def monitor(self):
        # set wait signal
        # if en == 1 take data
        intf = self.intf
        yield WaitWriteOnly()
        intf.wait.write(0)

        yield WaitCombStable()
        # wait for potential update of en

        en = intf.en.read()
        try:
            en = int(en)
        except ValueError:
            raise AssertionError(self.sim.now, intf,
                                 "en signal in invalid state")

        if en:
            yield Timer(CLK_PERIOD // 10)
            yield WaitCombRead()
            self.data.append(self.get_data())

    def driver(self):
        # if wait == 0 set en=1 and set data
        intf = self.intf

        d = None
        v = 0
        yield WaitCombRead()
        if self.notReset() and self.data:
            yield WaitCombRead()

            wait = intf.wait.read()
            try:
                wait = int(wait)
            except ValueError:
                raise AssertionError(self.sim.now, intf,
                                     "wait signal in invalid state")
            if not wait:
                d = self.data.popleft()
                v = 1

        yield WaitWriteOnly()
        self.set_data(d)
        intf.en.write(v)

    def getDrivers(self):
        return SyncAgentBase.getDrivers(self) + [self.driver_init()]

    def getMonitors(self):
        return SyncAgentBase.getMonitors(self) + [self.monitor_init()]



from typing import Union

from hwt.doc_markers import internal
from hwt.hdl.value import HValue
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
# from hwt.hdl.operator import Operator


@internal
def replace_input_in_expr(parentObj: Union["Operator", "HdlStatement"],
                          expr: Union[RtlSignalBase, HValue],
                          toReplace: RtlSignalBase,
                          replacement: RtlSignalBase,
                          updateEndpoints: bool) -> RtlSignalBase:
    """
    :return: True if expr is toReplace and should be replaced else False
    """
    if expr is toReplace:
        if updateEndpoints:
            expr.endpoints.discard(parentObj)
            replacement.endpoints.append(parentObj)
        return replacement
    elif isinstance(expr, RtlSignalBase) and expr.hidden:
        op = expr.origin
        # assert isinstance(op, Operator), op
        new_operands = []
        for o in op.operands:
            new_o = replace_input_in_expr(op, o, toReplace, replacement, True)
            new_operands.append(new_o)

        res = op.operator._evalFn(*new_operands)
        return res
    else:
        return expr



from hdlConvertorAst.hdlAst._bases import iHdlStatement
from hdlConvertorAst.hdlAst._expr import HdlAll
from hdlConvertorAst.hdlAst._statements import HdlStmProcess, HdlStmWait,\
    HdlStmBlock
from hdlConvertorAst.to.verilog.constants import SIGNAL_TYPE
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.codeBlockContainer import HdlStmCodeBlockContainer
from hwt.serializer.verilog.utils import verilogTypeOfSig


class ToHdlAstVerilog_statements():

    def as_hdl_HdlAssignmentContainer(self, a: HdlAssignmentContainer):
        blocking = False
        ver_sig_t = verilogTypeOfSig(a.dst)
        if ver_sig_t in (SIGNAL_TYPE.REG, SIGNAL_TYPE.PORT_REG):
            evDep = False
            for driver in a.dst.drivers:
                if driver._event_dependent_from_branch is not None:
                    evDep = True
                    break

            if not evDep or a.dst.virtual_only:
                blocking = True
        elif ver_sig_t in (SIGNAL_TYPE.WIRE, SIGNAL_TYPE.PORT_WIRE):
            blocking = True
        else:
            raise ValueError(ver_sig_t)

        a = super(ToHdlAstVerilog_statements, self).as_hdl_HdlAssignmentContainer(a)
        a.is_blocking = blocking
        return a

    def can_pop_process_wrap(self, stms, hasToBeVhdlProcess):
        if hasToBeVhdlProcess:
            return False
        else:
            assert len(stms) == 1
            return True

    def has_to_be_process(self, proc: HdlStmCodeBlockContainer):
        for o in proc._outputs:
            if verilogTypeOfSig(o) in (SIGNAL_TYPE.REG, SIGNAL_TYPE.PORT_REG):
                return True

        return False

    def as_hdl_HdlStmCodeBlockContainer(self, proc: HdlStmCodeBlockContainer) -> iHdlStatement:
        p = super(ToHdlAstVerilog_statements,
                  self).as_hdl_HdlStmCodeBlockContainer(proc)
        if isinstance(p, HdlStmProcess):
            no_wait = True
            if isinstance(p.body, HdlStmWait):
                no_wait = False
            elif isinstance(p.body, HdlStmBlock):
                for _o in p.body.body:
                    if isinstance(_o, HdlStmWait):
                        no_wait = False
                        break

            if no_wait and not p.sensitivity:
                # all input are constant and that is why this process does not have
                # any sensitivity
                p.sensitivity = [HdlAll, ]

            # add label
            if not isinstance(p.body, HdlStmBlock):
                b = p.body
                p.body = HdlStmBlock()
                p.body.body.append(b)
            p.body.labels.extend(p.labels)
            p.labels.clear()
        return p



from hwt.hdl.value import HValue, areHValues
from hwt.serializer.generic.indent import getIndent
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class StructValBase(HValue):
    """
    Base class for values for structure types.
    Every structure type has it's own value class derived from this.
    """
    __slots__ = []

    def __init__(self, typeObj, val, skipCheck=False):
        """
        :param val: None or dict {field name: field value}
        :param typeObj: instance of HString HdlType
        :param skipCheck: flag to skip field name consistency in val
        """
        self._dtype = typeObj
        if not skipCheck and val is not None:
            assert set(self.__slots__).issuperset(set(val.keys())), \
                ("struct value specifies undefined members",
                 set(val.keys()).difference(set(self.__slots__)))

        for f in self._dtype.fields:
            if f.name is None:
                continue
            if val is None:
                v = None
            else:
                v = val.get(f.name, None)

            if not isinstance(v, (HValue, RtlSignalBase)):
                v = f.dtype.from_py(v)

            setattr(self, f.name, v)

    def __copy__(self):
        d = {}
        for f in self._dtype.fields:
            if f.name is None:
                continue

            v = getattr(self, f.name)
            if not isinstance(v, RtlSignalBase):
                v = v.__copy__()
            d[f.name] = v

        return self.__class__(self._dtype, d, skipCheck=True)

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        """
        :param val: None or dict {field name: field value}
        :param typeObj: instance of HString HdlType
        :param vld_mask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        if vld_mask == 0:
            val = None
        return cls(typeObj, val)

    def _is_full_valid(self):
        for f in self._dtype.fields:
            if f.name is not None:
                val = getattr(self, f.name, None)
                if val is None or not val._is_full_valid():
                    return False
        return True

    def to_py(self):
        if not self._is_full_valid():
            raise ValueError(f"Value of {self} is not fully defined")
        d = {}
        for f in self._dtype.fields:
            if f.name is not None:
                val = getattr(self, f.name).to_py()
                d[f.name] = val
        return d

    def __ne__(self, other):
        if areHValues(self, other):
            if self._dtype == other._dtype:
                for f in self._dtype.fields:
                    isPadding = f.name is None
                    if not isPadding:
                        sf = getattr(self, f.name)
                        of = getattr(other, f.name)
                        if (sf != of):
                            return True
                return False
            else:
                return True
        else:
            return super(HValue, self).__ne__(other)

    def _eq(self, other):
        return self.__eq__(other)

    def __eq__(self, other):
        if areHValues(self, other):
            if self._dtype == other._dtype:
                for f in self._dtype.fields:
                    isPadding = f.name is None
                    if not isPadding:
                        sf = getattr(self, f.name)
                        of = getattr(other, f.name)
                        if not (sf == of):
                            return False
                return True
            else:
                return False
        else:
            return super(HValue, self).__eq__(other)

    def __repr__(self, indent=0):
        buff = ["{"]
        indentOfFields = getIndent(indent + 1)

        for f in self._dtype.fields:
            if f.name is not None:
                val = getattr(self, f.name)
                try:
                    v = val.__repr__(indent=indent + 1)
                except TypeError:
                    v = repr(val)

                buff.append(f"{indentOfFields:s}{f.name:s}: {v:s}")

        buff.append(getIndent(indent) + "}")
        return ("\n").join(buff)



from hwt.code import Concat
from hwt.doc_markers import internal
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import default_reinterpret_cast_fn, HdlType
from hwt.hdl.types.struct import HStruct
from hwt.hdl.value import HValue
from hwt.interfaces.std import Signal
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


@internal
def hstruct_reinterpret_to_bits(self, sigOrVal, toType: HdlType):
    assert toType.bit_length() == self.bit_length()
    parts = []
    for f in self.fields:
        if f.name is None:
            width = f.dtype.bit_length()
            part = Bits(width).from_py(None)
        else:
            part = getattr(sigOrVal, f.name)
            if isinstance(part, Signal):
                part = part._sig
            if not isinstance(part, (HValue, RtlSignalBase)):
                part = f.dtype.from_py(part)
            elif isinstance(part, HValue) and not isinstance(part._dtype, toType.__class__):
                part = part._reinterpret_cast(toType.__class__(part._dtype.bit_length()))

        parts.append(part)

    return Concat(*reversed(parts))


@internal
def hstruct_reinterpret_using_bits(self, sigOrVal, toType: HdlType):
    as_bits = sigOrVal._reinterpret_cast(Bits(self.bit_length()))
    return as_bits._reinterpret_cast(toType)


@internal
def hstruct_reinterpret(self, sigOrVal, toType: HdlType):
    if isinstance(toType, Bits):
        return hstruct_reinterpret_to_bits(self, sigOrVal, toType)
    elif isinstance(toType, (HStruct, HArray)):
        return hstruct_reinterpret_using_bits(self, sigOrVal, toType)
    else:
        return default_reinterpret_cast_fn(self, sigOrVal, toType)



from hwt.code import Concat
from hwt.doc_markers import internal
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.hdlType import default_reinterpret_cast_fn
from hwt.synthesizer.exceptions import TypeConversionErr
from pyMathBitPrecise.bit_utils import get_bit_range, mask
from hwt.hdl.types.struct import HStruct


@internal
def getBits_from_array(array, wordWidth, start, end,
                       reinterpretElmToType=None):
    """
    Gets value of bits between selected range from memory

    :param start: bit address of start of bit of bits
    :param end: bit address of first bit behind bits
    :return: instance of BitsVal (derived from SimBits type) which contains
        copy of selected bits
    """
    inPartOffset = 0
    value = Bits(end - start, None).from_py(None)

    while start != end:
        assert start < end, (start, end)

        dataWordIndex = start // wordWidth

        v = array[dataWordIndex]
        if reinterpretElmToType is not None:
            v = v._reinterpret_cast(reinterpretElmToType)

        endOfWord = (dataWordIndex + 1) * wordWidth
        width = min(end, endOfWord) - start
        offset = start % wordWidth

        val = get_bit_range(v.val, offset, width)
        vld_mask = get_bit_range(v.vld_mask, offset, width)

        m = mask(width)
        value.val |= (val & m) << inPartOffset
        value.vld_mask |= (vld_mask & m) << inPartOffset

        inPartOffset += width
        start += width

    return value


@internal
def reinterptet_harray_to_bits(typeFrom, sigOrVal, bitsT):
    """
    Cast HArray signal or value to signal or value of type Bits
    """
    size = int(typeFrom.size)
    widthOfElm = typeFrom.element_t.bit_length()
    w = bitsT.bit_length()
    if size * widthOfElm != w:
        raise TypeConversionErr(
            "Size of types is different", size * widthOfElm, w)

    partT = Bits(widthOfElm)
    parts = [p._reinterpret_cast(partT) for p in sigOrVal]

    return Concat(*reversed(parts))._reinterpret_cast(bitsT)


@internal
def reinterpret_harray_to_harray(typeFrom, sigOrVal, arrayT):
    mySize = int(typeFrom.size)
    myWidthOfElm = typeFrom.element_t.bit_length()
    size = int(arrayT.size)
    widthOfElm = arrayT.element_t.bit_length()

    if size * widthOfElm != mySize * myWidthOfElm:
        raise TypeConversionErr("Size of types is different",
                                size * widthOfElm, mySize * myWidthOfElm)

    if isinstance(typeFrom.element_t, Bits):
        reinterpretElmToType = None
    else:
        reinterpretElmToType = Bits(myWidthOfElm)

    res = arrayT.from_py(None)
    res.vld_mask = sigOrVal.vld_mask
    for i in range(size):
        start = i * widthOfElm
        end = (i + 1) * widthOfElm
        item = getBits_from_array(
            sigOrVal, myWidthOfElm, start, end, reinterpretElmToType)
        res[i] = item._reinterpret_cast(arrayT.element_t)

    return res


@internal
def reinterpret_harray_to_hstruct(typeFrom, sigOrVal, structT):
    as_bits = sigOrVal._reinterpret_cast(Bits(typeFrom.bit_length()))
    return as_bits._reinterpret_cast(structT)


@internal
def reinterpret_cast_harray(typeFrom, sigOrVal, toType):
    if isinstance(toType, Bits):
        return reinterptet_harray_to_bits(typeFrom, sigOrVal, toType)
    elif isinstance(toType, HArray):
        return reinterpret_harray_to_harray(typeFrom, sigOrVal, toType)
    elif isinstance(toType, HStruct):
        return reinterpret_harray_to_hstruct(typeFrom, sigOrVal, toType)
    else:
        return default_reinterpret_cast_fn(typeFrom, sigOrVal, toType)



from typing import Generator

from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.portItem import HdlPortItem
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.statements.statement import HdlStatement
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import SignalDriverErrType, \
    SignalDriverErr
from ipCorePackager.constants import DIRECTION


@internal
def walk_assignments(stm: HdlStatement, dst: RtlSignal)\
        ->Generator[HdlAssignmentContainer, None, None]:
    if isinstance(stm, HdlAssignmentContainer):
        if dst is stm.dst:
            yield stm
    else:
        for _stm in stm._iter_stms():
            yield from walk_assignments(_stm, dst)


@internal
def markVisibilityOfSignalsAndCheckDrivers(netlist: "RtlNetlist"):
    """
    * check if all signals are driven by something
    * mark signals with hidden = False if they are connecting statements
      or if they are external interface
    """
    signals = netlist.signals
    interfaceSignals = netlist.interfaces

    signals_with_driver_issue = []
    for sig in signals:
        #if isinstance(sig._nop_val, (RtlSignal, InterfaceBase)):
        #    sig._nop_val.hidden = False

        driver_cnt = len(sig.drivers)
        has_comb_driver = False
        if driver_cnt > 1:
            sig.hidden = False
            for d in sig.drivers:
                if not isinstance(d, Operator):
                    sig.hidden = False

                is_comb_driver = False

                if isinstance(d, HdlPortItem):
                    is_comb_driver = True
                elif d._event_dependent_from_branch is None:
                    for a in walk_assignments(d, sig):
                        if not a.indexes\
                                and a._event_dependent_from_branch != 0:
                            is_comb_driver = True
                            break

                if has_comb_driver and is_comb_driver:
                    signals_with_driver_issue.append(
                        (SignalDriverErrType.MULTIPLE_COMB_DRIVERS, sig))

                has_comb_driver |= is_comb_driver
        elif driver_cnt == 1:
            if not isinstance(sig.drivers[0], Operator):
                sig.hidden = False
        else:
            sig.hidden = False
            if sig not in interfaceSignals.keys():
                if not sig.def_val._is_full_valid():
                    signals_with_driver_issue.append(
                        (SignalDriverErrType.MISSING_DRIVER, sig))
                sig._const = True

        # chec interface direction if required
        d = interfaceSignals.get(sig, None)
        if d is None:
            pass
        elif d is DIRECTION.IN:
            assert sig.drivers, sig
            if len(sig.drivers) != 1:
                signals_with_driver_issue.append(
                    (SignalDriverErrType.INPUT_WITH_DRIVER, sig))
        elif d is DIRECTION.OUT:
            if not sig.drivers:
                signals_with_driver_issue.append(
                    (SignalDriverErrType.OUTPUT_WITHOUT_DRIVER, sig))

    if signals_with_driver_issue:
        raise SignalDriverErr(signals_with_driver_issue)



import re
from typing import List

from hdlConvertorAst.hdlAst import HdlOp, HdlModuleDec, HdlOpType, HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlAll
from hdlConvertorAst.hdlAst._statements import HdlImport, \
    HdlStmIf, HdlStmBlock, HdlStmFor, HdlStmForIn
from hdlConvertorAst.hdlAst._structural import HdlLibrary, HdlModuleDef, \
    HdlCompInst, HdlContext
from hdlConvertorAst.to.vhdl.keywords import VHLD2008_KEYWORDS
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword, NameScope
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call
from hwt.pyUtils.arrayQuery import groupedby
from hwt.serializer.generic.to_hdl_ast import ToHdlAst
from hwt.serializer.vhdl.ops import ToHdlAstVhdl2008_ops
from hwt.serializer.vhdl.statements import ToHdlAstVhdl2008_statements
from hwt.serializer.vhdl.types import ToHdlAstVhdl2008_types
from hwt.serializer.vhdl.value import ToHdlAstVhdl2008_Value


class VhdlNameScope(NameScope):
    RE_MANY_UNDERSCORES = re.compile(r"(_{2,})")

    def checked_name(self, actualName, actualObj):
        actualName = self.RE_MANY_UNDERSCORES.sub(r"_", actualName)
        if actualName[0] == "_":
            actualName = "u" + actualName
        return NameScope.checked_name(self, actualName, actualObj)


IEEE = HdlValueId("IEEE", obj=LanguageKeyword())
std_logic_1164 = HdlValueId("std_logic_1164", obj=LanguageKeyword())
numeric_std = HdlValueId("numeric_std", obj=LanguageKeyword())


class ToHdlAstVhdl2008(ToHdlAstVhdl2008_Value,
                       ToHdlAstVhdl2008_ops, ToHdlAstVhdl2008_types,
                       ToHdlAstVhdl2008_statements, ToHdlAst):
    """
    :ivar ~.name_scope: name scope used to generate a unique names for tmp variables
        (all object should be registered in namescope before serialization)
    """
    _keywords_dict = {kw: LanguageKeyword() for kw in VHLD2008_KEYWORDS}

    DEFAULT_IMPORTS = [
        HdlLibrary("IEEE"),
        HdlImport([IEEE, std_logic_1164, HdlAll]),
        HdlImport([IEEE, numeric_std, HdlAll]),
    ]

    @classmethod
    def getBaseNameScope(cls):
        s = VhdlNameScope.make_top(True)
        s.update(cls._keywords_dict)
        return s

    @staticmethod
    def _find_HdlCompInst(o):
        if isinstance(o, (list, tuple)):
            for _o in o:
                yield from ToHdlAstVhdl2008._find_HdlCompInst(_o)
        if isinstance(o, HdlCompInst):
            yield o
        elif isinstance(o, HdlStmBlock) and o.in_preproc:
            yield from ToHdlAstVhdl2008._find_HdlCompInst(o.body)
        elif isinstance(o, HdlStmIf) and o.in_preproc:
            if o.if_true:
                yield from ToHdlAstVhdl2008._find_HdlCompInst(o.if_true)
            for _, stms in o.elifs:
                yield from ToHdlAstVhdl2008._find_HdlCompInst(stms)
            if o.if_false:
                yield from ToHdlAstVhdl2008._find_HdlCompInst(o.if_false)
        elif isinstance(o, (HdlStmFor, HdlStmForIn)) and o.in_preproc:
            if o.body:
                yield from ToHdlAstVhdl2008._find_HdlCompInst(o.body)

    def _as_hdl_HdlModuleDef_param_asserts(self, new_m: HdlModuleDec) -> List[HdlOp]:
        res = []
        for p in new_m.params:
            p: HdlIdDef
            if p.value is None:
                continue
            a = hdl_call(HdlValueId("assert"), [
                 HdlOp(HdlOpType.EQ, [HdlValueId(p.name), self.as_hdl(p.value)]),
                 "Generated only for this value",
                 HdlValueId("error")])
            res.append(a)

        return res

    def as_hdl_HdlModuleDef(self, o: HdlModuleDef):
        """
        Translate hwt types and expressions to HDL AST and add explicit components
        """
        _o = super(ToHdlAstVhdl2008, self).as_hdl_HdlModuleDef(o)
        component_insts = []
        for c in _o.objs:
            component_insts.extend(self._find_HdlCompInst(c))

        # select comonent instances whith an unique module_name
        components = [
            x[1][0] for x in
            groupedby(component_insts, lambda c: c.module_name)
        ]
        components.sort(key=lambda c: c.module_name)
        components = [self.as_hdl_HldComponent(c)
                      for c in components]
        param_asserts = self._as_hdl_HdlModuleDef_param_asserts(o.dec)
        if components or param_asserts:
            # :note: it is important that the asserts are at the end because
            # we are detecting the declarations from the beginnig and assert there would
            # disturb that
            objs = [*components, *_o.objs, *param_asserts]
            _o.objs = objs

        res = HdlContext()
        res.objs.extend(self.DEFAULT_IMPORTS)
        res.objs.append(_o)
        return res

    def as_hdl_HldComponent(self, o: HdlCompInst):
        c = self.as_hdl_HdlModuleDec(o.origin._ctx.ent)
        return c



from hdlConvertorAst.hdlAst._defs import HdlIdDef
from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlOp, HdlOpType, \
    HdlTypeType
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_index, \
    hdl_downto
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, INT, FLOAT64
from hwt.hdl.types.float import HFloat


class ToHdlAstVhdl2008_types():
    BOOLEAN = HdlValueId("BOOLEAN", obj=LanguageKeyword())
    INTEGER = HdlValueId("INTEGER", obj=LanguageKeyword())
    STRING = HdlValueId("STRING", obj=LanguageKeyword())
    STD_LOGIC_VECTOR = HdlValueId("STD_LOGIC_VECTOR", obj=LanguageKeyword())
    STD_LOGIC = HdlValueId("STD_LOGIC", obj=LanguageKeyword())
    SIGNED = HdlValueId("SIGNED", obj=LanguageKeyword())
    UNSIGNED = HdlValueId("UNSIGNED", obj=LanguageKeyword())

    def as_hdl_HdlType_str(self, typ, declaration=False):
        assert not declaration
        return self.STRING

    def as_hdl_HdlType_bits(self, typ: Bits, declaration=False):
        if declaration:
            raise NotImplementedError()
        if typ == BOOL:
            return self.BOOLEAN
        if typ == INT:
            return self.INTEGER

        bitLength = typ.bit_length()
        w = typ.bit_length()
        isVector = typ.force_vector or bitLength > 1

        if typ.signed is None:
            if isVector:
                name = self.STD_LOGIC_VECTOR
            else:
                return self.STD_LOGIC
        elif typ.signed:
            name = self.SIGNED
        else:
            name = self.UNSIGNED

        return HdlOp(HdlOpType.CALL, [
            name,
            HdlOp(HdlOpType.DOWNTO, [
                self.as_hdl(w - 1),
                self.as_hdl_int(0)
            ])])

    def as_hdl_HdlType_array(self, typ: HArray, declaration=False):
        if declaration:
            v = HdlIdDef()
            name = getattr(typ, "name", None)
            if name is None:
                name = "arr_t_"
            v.name = self.name_scope.checked_name(name, typ)
            v.type = HdlTypeType
            v.origin = typ
            size = hdl_downto(
                self.as_hdl_int(int(typ.size) - 1),
                self.as_hdl_int(0)
            )
            if self.does_type_requires_extra_def(typ.element_t, ()):
                raise NotImplementedError(typ.element_t)
            e_t = self.as_hdl_HdlType(typ.element_t, declaration=False)
            v.value = hdl_index(e_t, size)
            return v
        else:
            return super(ToHdlAstVhdl2008_types, self).as_hdl_HdlType_array(typ, declaration)

    def as_hdl_HdlType_float(self, typ: HFloat, declaration=False):
        if typ == FLOAT64:
            return HdlValueId("real")
        else:
            raise NotImplementedError(typ)



from copy import copy
import math
from typing import List, Union

from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bitsVal import BitsVal
from hwt.hdl.value import HValue
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from pyMathBitPrecise.bit_utils import mask


def inRange(n, start, end):
    """
    Check if n is in range <start, end)
    """
    res = (n >= start)
    if isinstance(n, (RtlSignalBase, InterfaceBase, HValue)) and\
            (not isinstance(end, int) or
             end < 2 ** n._dtype.bit_length()):
        res = res & (n < end)
    return res


def toPow2Ceil(x: int):
    """
    Get the smallest 2**N where 2**N >= x
    """
    i = 0
    while 2 ** i < x:
        i += 1
    return 2 ** i


def addressAlignBestEffort(record_width: int, bus_data_width: int):
    """
    Optionally extend the record width to be power of 2 and to consume
    smallest amount of memory possible.
    """
    if 2 * record_width <= bus_data_width:
        # multiple records in bus data word
        records_in_bus_word = math.floor(bus_data_width / record_width)
        record_width = bus_data_width // records_in_bus_word
        bus_words_in_record = 1
    else:
        # record in a 1+ bus data words
        bus_words_in_record = math.ceil(record_width / bus_data_width)
        # in order to make item indexable we need to have size which is 2**X
        bus_words_in_record = toPow2Ceil(bus_words_in_record)

        record_width = bus_data_width * bus_words_in_record
        records_in_bus_word = 1

    return record_width, records_in_bus_word, bus_words_in_record


def log2ceil(x):
    """
    Returns no of bits required to store x-1
    for example x=8 returns 3
    """

    if not isinstance(x, (int, float)):
        x = int(x)

    if x == 0 or x == 1:
        res = 1
    else:
        res = math.ceil(math.log2(x))

    return res


def isPow2(num) -> bool:
    """
    Check if number or constant is power of two
    """
    if not isinstance(num, int):
        num = int(num)
    return num != 0 and ((num & (num - 1)) == 0)


def sizeof(_type) -> int:
    "get size of type in bytes"
    s = _type.bit_length()
    return math.ceil(s / 8)


def shiftIntArray(values: List[Union[int, BitsVal]], item_width: int, shift: int):
    """
    :param values: array of values which will be shifted as a whole
    :param item_width: a bit length of a single item in array
    :param shift: specifies how many bits the array should be shifted, << is a positive shift, >> is a negative shift
    """
    if shift == 0:
        return copy(values)
    new_v = []
    t = Bits(item_width)
    if shift > 0:
        # <<
        for _ in range(shift // item_width):
            new_v.append(None)
        prev = None
        for v in values:
            if v is None and prev is None:
                _v = None
            else:
                if prev is None:
                    prev = t.from_py(None)
                if v is None:
                    v = t.from_py(None)
                elif isinstance(prev, BitsVal) and not isinstance(v, BitsVal):
                    v = t.from_py(v)
                _v = (v << shift) | (prev >> (item_width - shift))
            new_v.append(_v)
            prev = v
        if prev is None:
            v = None
        else:
            v = prev >> (item_width - shift)
        new_v.append(v)
    else:
        # shift < 0, >>
        nextIt = iter(values)
        for v in values:
            try:
                nv = next(nextIt)
            except StopIteration:
                nv = None
            if nv is None and v is None:
                _v = None
            else:
                if nv is None:
                    nv = t.from_py(None)
                if v is None:
                    v = t.from_py(None)
                _v = (v >> shift) | (nv & mask(shift))
            new_v.append(_v)

    return new_v



from typing import List, Set

from hwt.doc_markers import internal
from hwt.hdl.statements.statement import HdlStatement
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


@internal
def HdlStatement_discover_enclosure_for_statements(
        statements: List[HdlStatement],
        outputs: List[RtlSignalBase]) -> Set[RtlSignalBase]:
    """
    Discover enclosure for list of statements

    :param statements: list of statements in one code branch
    :param outputs: list of outputs which should be driven from this statement list
    :return: set of signals for which this statement list have always some driver
        (is enclosed)
    """
    result = set()
    if not statements:
        return result

    for stm in statements:
        stm._discover_enclosure()

    for o in outputs:
        has_driver = False

        for stm in statements:
            if o in stm._outputs:
                assert not has_driver
                has_driver = False
                result.update(stm._enclosed_for)
            else:
                pass

    return result




from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL, INT
from hwt.hdl.types.slice import HSlice
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import HValue
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from pyMathBitPrecise.bit_utils import ValidityError


class HArrayVal(HValue):
    """
    Class for values of array HDL type
    """

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        """
        :param val: None or dictionary {index:HValue} or iterrable of values
        :param vld_mask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        size = typeObj.size
        if isinstance(size, HValue):
            size = int(size)

        elements = {}
        if vld_mask == 0:
            val = None

        if val is None:
            vld_mask = 0
        elif isinstance(val, dict):
            if vld_mask is None:
                vld_mask = 1

            for k, v in val.items():
                if not isinstance(k, int):
                    k = int(k)
                e = elements[k] = typeObj.element_t.from_py(v)
                vld_mask &= e._is_full_valid()
        else:
            if vld_mask is None:
                vld_mask = 1

            for k, v in enumerate(val):
                if isinstance(v, RtlSignalBase):  # is signal
                    assert v._dtype == typeObj.element_t
                    e = v
                else:
                    e = typeObj.element_t.from_py(v)
                elements[k] = e
                vld_mask &= e._is_full_valid()

        if len(elements) != size:
            vld_mask = 0

        return cls(typeObj, elements, vld_mask)

    def to_py(self):
        if not self._is_full_valid():
            raise ValidityError(f"Value of {self} is not fully defined")

        v = self.val
        invalid_elm = self._dtype.element_t.from_py(None)
        return [v.get(i, invalid_elm).to_py()
                for i in range(self._dtype.size)]

    @internal
    def __hash__(self):
        return hash(self._dtype)
        # return hash((self._dtype, self.val, self.vld_mask))

    def _is_full_valid(self):
        return self.vld_mask == 1

    @internal
    def _getitem__val(self, key):
        """
        :atention: this will clone item from array, iterate over .val
            if you need to modify items
        """
        try:
            kv = key.val
            if not key._is_full_valid():
                raise KeyError()
            else:
                if kv >= self._dtype.size:
                    raise KeyError()

            return self.val[kv].__copy__()
        except KeyError:
            return self._dtype.element_t.from_py(None)

    def __getitem__(self, key):
        iamVal = isinstance(self, HValue)
        key = toHVal(key)
        isSLICE = isinstance(key, HSlice.getValueCls())

        if isSLICE:
            raise NotImplementedError()
        elif isinstance(key, (HValue, RtlSignalBase)):
            pass
        else:
            raise NotImplementedError(
                f"Index operation not implemented for index {key}")

        if iamVal and isinstance(key, HValue):
            return self._getitem__val(key)

        return Operator.withRes(AllOps.INDEX, [self, key], self._dtype.element_t)

    @internal
    def _setitem__val(self, index, value):
        if index._is_full_valid():
            self.val[index.val] = value.__copy__()
        else:
            self.val = {}

    def __setitem__(self, index, value):
        """
        Only syntax sugar for user, not used inside HWT

        * Not used in HW design (__getitem__ and overloaded call operator is used instead for item assigning)
        * In simulator _setitem__val is used directly
        """
        if isinstance(index, int):
            index = INT.from_py(index)
        else:
            assert isinstance(self, HValue)
            assert isinstance(index._dtype, Bits), index._dtype

        if not isinstance(value, HValue):
            value = self._dtype.element_t.from_py(value)
        else:
            assert value._dtype == self._dtype.element_t, (
                value._dtype, self._dtype.element_t)

        ret = self._setitem__val(index, value)
        self.vld_mask = int(
            len(self.val) == self._dtype.size and
            all(e._is_full_valid() for e in self.val.values())
        )
        return ret

    def __iter__(self):
        mySize = len(self)

        for i in range(mySize):
            yield self[i]

    def __len__(self):
        return int(self._dtype.size)

    @internal
    def _eq__val(self, other):
        assert self._dtype.element_t == other._dtype.element_t
        assert self._dtype.size == other._dtype.size

        eq = True
        vld = 1
        keysA = set(self.val)
        keysB = set(other.val)
        sharedKeys = keysA.union(keysB)

        lsh = len(sharedKeys)
        if (lsh == int(self._dtype.size)
                and len(keysA) == lsh
                and len(keysB) == lsh):
            for k in sharedKeys:
                a = self.val[k]
                b = other.val[k]

                eq = eq and bool(a) == bool(b)
                if not eq:
                    break
                vld = vld & a.vld_mask & b.vld_mask
        else:
            eq = False
            vld = 0

        return BOOL.getValueCls()(BOOL, int(eq), vld)

    def _eq(self, other):
        assert isinstance(other, HArrayVal)
        return self._eq__val(other)



"""
Verilog serializer serializes HDL objects to verilog code.
"""
from hdlConvertorAst.to.verilog.verilog2005 import ToVerilog2005
from hwt.serializer.serializer_config import DummySerializerConfig
from hwt.serializer.verilog.serializer import ToHdlAstVerilog
from hwt.serializer.xdc.serializer import XdcSerializer


class VerilogSerializer(DummySerializerConfig):
    fileExtension = '.v'
    TO_HDL_AST = ToHdlAstVerilog
    TO_HDL = ToVerilog2005
    TO_CONSTRAINTS = XdcSerializer


from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwtSimApi.agents.base import AgentBase, SyncAgentBase as pcSyncAgentBase,\
    AgentWitReset as pcAgentWitReset
from hwtSimApi.hdlSimulator import HdlSimulator
from hwtSimApi.process_utils import OnRisingCallbackLoop


class AgentWitReset(pcAgentWitReset):

    def __init__(self, sim: HdlSimulator, intf, allowNoReset=False):
        pcAgentWitReset.__init__(self, sim, intf, (None, False))
        self.rst, self.rstOffIn = self._discoverReset(intf, allowNoReset)

    @classmethod
    def _discoverReset(cls, intf, allowNoReset):
        try:
            rst = intf._getAssociatedRst()
            rstOffIn = int(rst._dtype.negated)
            rst = rst._sigInside
        except IntfLvlConfErr:
            rst = None
            rstOffIn = True
            if not allowNoReset:
                raise
        return (rst, rstOffIn)

    def notReset(self):
        if self.rst is None:
            return True
        else:
            rstVal = self.rst.read()
            rstVal = int(rstVal)
            return rstVal == self.rstOffIn


class SyncAgentBase(AgentWitReset, pcSyncAgentBase):
    """
    Agent which discovers clk, rst signal and runs only
    at specified edge of clk

    :attention: requires clk and rst/rstn signal
        (if you do not have any create simulation wrapper with it)
    """
    SELECTED_EDGE_CALLBACK = OnRisingCallbackLoop

    def __init__(self, sim: HdlSimulator, intf, allowNoReset=False,
                 wrap_monitor_and_driver_in_edge_callback=True):
        self.intf = intf
        clk = self.intf._getAssociatedClk()
        rst = self._discoverReset(intf, allowNoReset)
        pcSyncAgentBase.__init__(
            self, sim, intf, clk, rst, 
            wrap_monitor_and_driver_in_edge_callback=wrap_monitor_and_driver_in_edge_callback)





class ResourceError(Exception):
    """
    An error which means that the resource of this kind
    does not exists in current hardware.
    """


class ResourceMUX():
    def __init__(self, bitWidth, inputs):
        self.bitWidth = bitWidth
        self.inputs = inputs

    def __repr__(self):
        return f"<{self.__class__.__name__:s} {self.bitWidth:d} bits, {self.inputs} inputs>"

class ResourceFF():
    pass


class ResourceLatch():
    pass


class ResourceRAM():
    """
    Specifier of type of RAM like memory
    """

    def __init__(self, width, items,
                 rwSync: int, rSync: int, wSync: int, rSync_wAsync: int,
                 rwAsync: int, rAsync: int, wAsync: int, rAsync_wSync: int):
        """
        :param width: widtho of word in RAM/ROM
        :param items: number of words in RAM/ROM
        :param rwSync: count of read + write sychronous ports
        :param rSync: count of read only sychronous ports
        :param wSync: count of write only sychronous ports
        :param rSync_wAsync: count of synchronous read + asynchronous write ports
        :param rwAsync: count of read + write asychronous ports
        :param rAsync: count of read only asychronous ports
        :param wAsync: count of write only asychronous ports
        :param rAsync_wSync: count of asynchronous read + synchronous write ports
        """
        self.width = width
        self.items = items
        self.rwSync = rwSync
        self.rSync = rSync
        self.wSync = wSync
        self.rSync_wAsync = rSync_wAsync
        self.rwAsync = rwAsync
        self.rAsync = rAsync
        self.wAsync = wAsync
        self.rAsync_wSync = rAsync_wSync

    def __hash__(self):
        return hash((
            self.width, self.items,
            self.rwSync, self.rSync, self.wSync,
            self.rSync_wAsync,
            self.rwAsync, self.rAsync, self.wAsync,
            self.rAsync_wSync))

    def __eq__(self, other):
        return isinstance(other, ResourceRAM) and (
            self.width == other.width and
            self.items == other.items and
            self.rwSync == other.rwSync and
            self.rSync == other.rSync and
            self.wSync == other.wSync and
            self.rSync_wAsync == other.rSync_wAsync and
            self.rwAsync == other.rwAsync and
            self.rAsync == other.rAsync and
            self.wAsync == other.wAsync and
            self.rAsync_wSync == other.rAsync_wSync
        )

    def __repr__(self):
        return ("<%s, %dbit x %d, syncPorts: (rw:%d, r:%d, w:%d), asyncPorts:"
                " (rw:%d, r:%d, w:%d), rSync_wAsyncPorts: %d,"
                " rAsync_wSyncPorts: %d>") % (
            self.__class__.__name__, self.width, self.items,
            self.rwSync, self.rSync, self.wSync,
            self.rwAsync, self.rAsync, self.wAsync,
            self.rSync_wAsync, self.rAsync_wSync
        )




from typing import Optional, Any

from hwt.hdl.types.defs import INT, STR, BOOL, SLICE, FLOAT64
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase

defaultPyConversions = {
    int: INT,
    str: STR,
    bool: BOOL,
    slice: SLICE,
    float: FLOAT64
}


def toHVal(op: Any, suggestedType: Optional[HdlType]=None):
    """Convert python or hdl value/signal object to hdl value/signal object"""
    if isinstance(op, HValue) or isinstance(op, SignalItem):
        return op
    elif isinstance(op, InterfaceBase):
        return op._sig
    else:
        if suggestedType is not None:
            return suggestedType.from_py(op)

        if isinstance(op, int):
            if op >= 1 << 31:
                raise TypeError(
                    f"Number {op:d} is too big to fit in 32 bit integer of HDL"
                    " use Bits type instead")
            elif op < -(1 << 31):
                raise TypeError(
                    f"Number {op:d} is too small to fit in 32 bit integer"
                    " of HDL use Bits type instead")

        try:
            hType = defaultPyConversions[type(op)]
        except KeyError:
            hType = None

        if hType is None:
            raise TypeError(f"Unknown hardware type for instance of {op.__class__}")

        return hType.from_py(op)



from hwt.doc_markers import internal
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.structValBase import StructValBase
from hwt.serializer.generic.indent import getIndent


class HStructFieldMeta():
    """
    Meta for field in struct type

    :ivar ~.split: flag which specifies if structured data type of this field
        should be synchronized as a one interface
        or each it's part should be synchronized separately
    """
    def __init__(self, split=False):
        self.split = split

    def __eq__(self, other):
        if other is None:
            return False
        return self.split == other.split

    @internal
    def __hash__(self):
        return hash(self.split)


class HStructField(object):
    def __init__(self, typ: HdlType, name: str, meta=None):
        assert isinstance(name, str) or name is None, name
        assert isinstance(typ, HdlType), typ
        self.name = name
        self.dtype = typ
        self.meta = meta

    def __eq__(self, other):
        return self.name == other.name and\
               self.dtype == other.dtype and\
               self.meta == other.meta

    def __hash__(self):
        return hash((self.name, self.dtype, self.meta))

    def __repr__(self):
        name = self.name
        if name is None:
            name = "<padding>"

        return f"<HStructField {self.dtype}, {name:s}>"


protectedNames = {"clone", "staticEval", "from_py", "_dtype"}


class HStruct(HdlType):
    """
    HDL structure type

    :ivar ~.fields: tuple of :class:`~.HStructField` instances in this struct
    :ivar ~.name: name of this HStruct type
    :ivar ~.field_by_name: dictionary which maps the name of the field to :class:`~.HStructField` instance
    :ivar ~.valueCls: Class of value for this type as usual
        in HdlType implementations
    """
    def __init__(self, *template, name=None, const=False):
        """
        :param template: list of tuples (type, name) or :class:`~.HStructField` objects
            name can be None (= padding)
        :param name: optional name used for debugging purposes
        """
        super(HStruct, self).__init__(const=const)

        fields = []
        field_by_name = {}
        self.name = name
        bit_length = 0
        for f in template:
            try:
                field = HStructField(*f)
            except TypeError:
                field = f
            if not isinstance(field, HStructField):
                raise TypeError(f"Template for struct field {f} is"
                                " not in valid format")

            fields.append(field)
            if field.name is not None:
                assert field.name not in field_by_name, field.name
                field_by_name[field.name] = field

            t = field.dtype
            if bit_length is not None:
                try:
                    _bit_length = t.bit_length()
                    bit_length += _bit_length
                except TypeError:
                    bit_length = None

        self.fields = tuple(fields)
        self.field_by_name = field_by_name
        self.__hash = hash((self.name, self.const, self.fields))
        self.__bit_length_val = bit_length

        usedNames = set(field_by_name.keys())
        assert not protectedNames.intersection(usedNames),\
            protectedNames.intersection(usedNames)

        class StructVal(StructValBase):
            __slots__ = list(usedNames)

        if name is not None:
            StructVal.__name__ = name + "Val"

        self.valueCls = StructVal

    def bit_length(self):
        bl = self.__bit_length_val
        if bl is None:
            raise TypeError("Can not request bit_lenght on type"
                            " which has not fixed size")
        else:
            return self.__bit_length_val

    @internal
    def getValueCls(self):
        return self.valueCls

    @internal
    @classmethod
    def get_reinterpret_cast_fn(cls):
        from hwt.hdl.types.structCast import hstruct_reinterpret
        return hstruct_reinterpret

    @internal
    def __fields__eq__(self, other):
        if len(self.fields) != len(other.fields):
            return False
        for sf, of in zip(self.fields, other.fields):
            if (sf.name != of.name or
                    sf.dtype != of.dtype or
                    sf.meta != of.meta):
                return False
        return True

    def __eq__(self, other):
        if self is other:
            return True
        if (type(self) is type(other)):
            if self.name != other.name or self.const != other.const:
                return False
            try:
                self_l = self.bit_length()
            except TypeError:
                self_l = -1
            try:
                other_l = other.bit_length()
            except TypeError:
                other_l = -1

            return self_l == other_l and self.__fields__eq__(other)
        return False

    @internal
    def __hash__(self):
        return self.__hash

    def __add__(self, other):
        """
        override of addition, merge struct into one
        """
        assert isinstance(other, HStruct)
        return HStruct(*self.fields, *other.fields)

    def __repr__(self, indent=0, withAddr=None, expandStructs=False):
        """
        :param indent: number of indentation
        :param withAddr: if is not None is used as a additional
            information about on which address this type is stored
            (used only by HStruct)
        :param expandStructs: expand HStructTypes (used by HStruct and HArray)
        """
        if self.name:
            name = self.name + " "
        else:
            name = ""

        myIndent = getIndent(indent)
        childIndent = getIndent(indent + 1)
        header = f"{myIndent:s}struct {name:s}{{"

        buff = [header, ]
        for f in self.fields:
            if withAddr is not None:
                withAddr_B = withAddr // 8
                addrTag = f" // start:0x{withAddr:x}(bit) 0x{withAddr_B:x}(byte)"
            else:
                addrTag = ""

            if f.name is None:
                buff.append(f"{childIndent:s}//{f.dtype} empty space{addrTag:s}")
            else:
                buff.append("%s %s%s" % (
                               f.dtype.__repr__(indent=indent + 1,
                                                withAddr=withAddr,
                                                expandStructs=expandStructs),
                            f.name, addrTag))
            if withAddr is not None:
                withAddr += f.dtype.bit_length()

        buff.append(f"{myIndent:s}}}")
        return "\n".join(buff)





def internal(fn):
    """
    Decorator which does not affect functionality but it is used as marker
    which tells that this object is not interesting for users and it is only used internally 
    """
    return fn


from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from typing import Set


class TypeConversionErr(TypeError):
    pass


class ConfErr(Exception):
    pass


class IntfLvlConfErr(ConfErr):
    """
    Interface level synthesizer user configuration error
    """
    pass


class SigLvlConfErr(ConfErr):
    """
    Signal level synthesizer user configuration error
    """
    pass


class InterfaceStructureErr(IntfLvlConfErr):
    """
    An exception which means that the two interfaces have non compatible sub-interfaces.
    (E.g. they do have a differently named signals)

    :ivar exclude: a set of sub-interfaces which should be excluded during the comparison
    """

    def __init__(self, dst: InterfaceBase, src: InterfaceBase, exclude: Set[InterfaceBase]):
        super(InterfaceStructureErr, self).__init__()
        self.src = src
        self.dst = dst
        self.exclude = exclude

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        missing_on_src = []
        missing_on_dst = []
        dst = self.dst
        src = self.src
        exclude = self.exclude
        for i in src._interfaces:
            i2 = getattr(dst, i._name, None)
            if i2 is None and (not exclude or i not in exclude):
                missing_on_dst.append(i._name)

        for i in dst._interfaces:
            i2 = getattr(src, i._name, None)
            if i2 is None and (not exclude or i not in exclude):
                missing_on_src.append(i._name)

        buff = [f"<{self.__class__.__name__} {dst} <= {src}"]
        if missing_on_dst:
            buff.append(f", missing on dst: {missing_on_dst}")
        if missing_on_src:
            buff.append(f", missing on src: {missing_on_src}")
        buff.append(">")
        return "".join(buff)

    def __copy__(self):
        return self.__class__(self.dst, self.src, self.exclude)



"""
VHDL serializer serializes HDL objects to VHDL code.
"""
from hdlConvertorAst.to.vhdl.vhdl2008 import ToVhdl2008
from hwt.serializer.vhdl.serializer import ToHdlAstVhdl2008
from hwt.serializer.xdc.serializer import XdcSerializer


class Vhdl2008Serializer():
    fileExtension = '.vhd'
    TO_HDL_AST = ToHdlAstVhdl2008
    TO_HDL = ToVhdl2008
    TO_CONSTRAINTS = XdcSerializer



from typing import Union

from hwt.hdl.operator import Operator
from hwt.hdl.statements.statement import HdlStatement
from hwt.serializer.resourceAnalyzer.resourceTypes import \
    ResourceFF, ResourceMUX, ResourceLatch, ResourceRAM
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.unit import Unit


class ResourceContext():
    """
    Container of informations about resources used in architecture

    :ivar ~.unit: optional unit for which is this context build
    :ivar ~.seen: set of seen objects
    :ivar ~.resources: dictionary {type of resource: cnt}
    :ivar ~.discoveredRamSignals: set of signals which seems to be some kind
        of RAM/ROM memory
    """

    def __init__(self, unit: Unit):
        self.unit = unit
        self.seen = set()
        self.resources = {}
        # {RtlSignal or Statement or Operator: HwResource or typle (HwResource, cnt)}
        self.resource_for_object = {}
        # {(mem, addr, syncFlag, r/w): cnt}
        self.memories = {}

    def registerOperator(self, op: Operator):
        w = op.operands[0]._dtype.bit_length()
        res = self.resources
        k = (op.operator, w)
        res[k] = res.get(k, 0) + 1

        self.resource_for_object[op] = k

    def registerMUX(self, stm: Union[HdlStatement, Operator], sig: RtlSignal,
                    inputs_cnt: int):
        """
        mux record is in format (self.MUX, n, m)
        where n is number of bits of this mux
        and m is number of possible inputs
        """
        assert inputs_cnt > 1
        res = self.resources
        w = sig._dtype.bit_length()
        k = (ResourceMUX, w, inputs_cnt)
        res[k] = res.get(k, 0) + 1

        self.resource_for_object[(stm, sig)] = k

    def registerFF(self, ff):
        res = self.resources
        w = ff._dtype.bit_length()
        ffs = res.get(ResourceFF, 0)
        res[ResourceFF] = ffs + w

        self.resource_for_object[ff] = (ResourceFF, w)

    def registerLatch(self, latch: RtlSignal):
        res = self.resources
        w = latch._dtype.bit_length()
        latches = res.get(ResourceLatch, 0)
        res[ResourceLatch] = latches + w

        self.resource_for_object[latch] = (ResourceLatch, w)

    def registerRAM_write_port(self, mem: RtlSignal, addr: RtlSignal,
                               synchronous: bool):
        res = self.memories
        addresses = res.get(mem, {})
        res[mem] = addresses

        # [rSycn, wSync, rAsync, wAsync]
        portCnts = addresses.get(addr, [0, 0, 0, 0])
        if synchronous:
            portCnts[1] += 1
        else:
            portCnts[3] += 1
        addresses[addr] = portCnts

    def registerRAM_read_port(self, mem: RtlSignal, addr: RtlSignal,
                              synchronous: bool):
        res = self.memories
        addresses = res.get(mem, {})
        res[mem] = addresses

        # [rSycn, wSync, rAsync, wAsync]
        portCnts = addresses.get(addr, [0, 0, 0, 0])
        if synchronous:
            portCnts[0] += 1
        else:
            portCnts[2] += 1
        addresses[addr] = portCnts

    def finalize(self):
        """
        Resolve ports of discovered memories
        """
        ff_to_remove = 0
        res = self.resources
        for m, addrDict in self.memories.items():
            rwSyncPorts, rSyncPorts, wSyncPorts = 0, 0, 0
            rwAsyncPorts, rAsyncPorts, wAsyncPorts = 0, 0, 0
            rSync_wAsyncPorts, rAsync_wSyncPorts = 0, 0

            for _, (rSync, wSync, rAsync, wAsync) in addrDict.items():
                if rSync:
                    ff_to_remove += rSync * m._dtype.element_t.bit_length()

                # resolve port count for this addr signal
                rwSync = min(rSync, wSync)
                rSync -= rwSync
                wSync -= rwSync

                rwAsync = min(rAsync, wAsync)
                rAsync -= rwAsync
                wAsync -= rwAsync

                rSync_wAsync = min(rSync, wAsync)
                rSync -= rSync_wAsync
                wAsync -= rSync_wAsync

                rAsync_wSync = min(rAsync, wSync)
                rAsync -= rAsync_wSync
                wSync -= rAsync_wSync

                # update port counts for mem
                rwSyncPorts += rwSync
                rSyncPorts += rSync
                wSyncPorts += wSync
                rwAsyncPorts += rwAsync
                rAsyncPorts += rAsync
                wAsyncPorts += wAsync

                rSync_wAsyncPorts += rSync_wAsync
                rAsync_wSyncPorts += rAsync_wSync
            k = ResourceRAM(m._dtype.element_t.bit_length(),
                            int(m._dtype.size),
                            rwSyncPorts, rSyncPorts, wSyncPorts,
                            rSync_wAsyncPorts,
                            rwAsyncPorts, rAsyncPorts, wAsyncPorts,
                            rAsync_wSyncPorts)
            res[k] = res.get(k, 0) + 1

        self.memories.clear()

        # remove register on read ports which will be merged into ram
        if ff_to_remove:
            ff_cnt = res[ResourceFF]
            ff_cnt -= ff_to_remove
            if ff_cnt:
                res[ResourceFF] = ff_cnt
            else:
                del res[ResourceFF]



from copy import deepcopy

from hwt.synthesizer.interface import Interface
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.unit import Unit


def to_tuple_of_names(objs):
    res = []
    for o in objs:
        if isinstance(o, RtlSignal):
            name = o.name
        elif hasattr(o, "_name"):
            name = o._name
        else:
            name = repr(o)
        res.append(name)
    return tuple(res)


class ComponentPath(tuple):

    def __new__ (cls, *objs):
        return super(ComponentPath, cls).__new__(cls, objs)

    def __truediv__(self, other):
        if isinstance(other, ComponentPath):
            return ComponentPath(*self, *other)
        else:
            return ComponentPath(*self, other)

    def resolve(self) -> "ComponentPath":
        """
        Make the path absolute

        The ComponentPath is in absolute format only if:

        * The first member is a top component or path is empty
        * All members except the last are :class:`hwt.synthesizer.unit.Unit` instances (last can be RtlSignal/Interface)
        * Each successor member is instanciated in predecessor except for :class:`hwt.synthesizer.unit.Unit` instance with shared component
        * If member is a :class:`hwt.synthesizer.unit.Unit` instance with shared component the successor must be an interface of this instance or an object from shared component
        """
        it = iter(reversed(self))
        try:
            obj = next(it)
        except StopIteration:
            # empty path
            return self

        path = []
        while obj is not None:
            _handle = next(it, None)
            if isinstance(_handle, Unit) and _handle._shared_component_with is not None:
                handle, _, _ = _handle._shared_component_with
            else:
                handle = _handle

            if obj is not handle:
                if isinstance(obj, RtlSignal):
                    # to not modify path if it is already in absolute format
                    if not path or path[-1] is not obj:
                        path.append(obj)
                    obj = obj.ctx.parent

                while isinstance(obj, Interface):
                    if not path:
                        path.append(obj)
                    if obj is handle:
                        break
                    obj = obj._parent

                while obj is not handle:
                    try:
                        assert isinstance(obj, Unit), obj
                    except:
                        raise
                    # to not modify path if it is already in absolute format
                    if not path or path[-1] is not obj:
                        path.append(obj)

                    obj = obj._parent

            obj = _handle

        return ComponentPath(*reversed(path))

    def is_absolute(self):
        """
        :return: True if path starts with a top component else False
        """
        u = self[0]
        return isinstance(u, Unit) and u._parent is None

    def update_prefix(self, old_path_prefix: "ComponentPath", new_path_prefix: "ComponentPath"):
        """
        Update prefix of the path tuple
        """
        assert len(self) >= len(old_path_prefix), (self, old_path_prefix)
        for p, op in zip(self, old_path_prefix):
            assert p is op, (self, old_path_prefix)

        return ComponentPath(*new_path_prefix, *self[len(old_path_prefix):])

    def __getitem__(self, key):
        if isinstance(key, slice):
            return self.__class__(*tuple.__getitem__(self, key))
        else:
            return tuple.__getitem__(self, key)

    def __repr__(self):
        return f"<{self.__class__.__name__:s} {str(self):s}>"

    def __str__(self):
        return "/".join(to_tuple_of_names(self))

    def __copy__(self):
        return self.__class__(*self)

    def __deepcopy__(self, memo):
        res = self.__class__(*(deepcopy(x, memo) for x in self))
        memo[self] = res
        return res



from hwt.hdl.constants import DIRECTION
from hwt.hdl.types.bits import Bits
from hwt.interfaces.std import Signal, Clk
from hwt.synthesizer.interface import Interface
from hwt.synthesizer.param import Param
from ipCorePackager.intfIpMeta import IntfIpMetaNotSpecified
from hwtSimApi.agents.peripheral.tristate import TristateAgent, TristateClkAgent
from hwtSimApi.hdlSimulator import HdlSimulator


class TristateSig(Interface):
    """
    Tristate interface

    :ivar ~.force_vector: in order to make this a vector[0] instead of single bit
        use FORCE_VECTOR=True
    """

    def _config(self):
        self.DATA_WIDTH = Param(1)
        self.FORCE_VECTOR = False

    def _declr(self):
        t = Bits(self.DATA_WIDTH, force_vector=self.FORCE_VECTOR)

        # connect
        self.t = Signal(dtype=t)
        # input
        self.i = Signal(dtype=t, masterDir=DIRECTION.IN)
        # output
        self.o = Signal(dtype=t)

    def _initSimAgent(self, sim: HdlSimulator):
        rst = self._getAssociatedRst()
        self._ag = TristateAgent(sim, self, (rst, rst._dtype.negated))


class TristateClk(Clk, TristateSig):
    def _config(self):
        Clk._config(self)
        TristateSig._config(self)

    def _getIpCoreIntfClass(self):
        raise IntfIpMetaNotSpecified()

    def _initSimAgent(self, sim: HdlSimulator):
        self._ag = TristateClkAgent(sim, self)



"""
Definitions of most common types
"""

from hwt.hdl.types.bits import Bits
from hwt.hdl.types.slice import HSlice
from hwt.hdl.types.string import HString
from hwt.hdl.types.float import HFloat


BOOL = Bits(bit_length=1, name="bool")
INT = Bits(bit_length=32, signed=True, name="int",
           strict_sign=False, strict_width=False)
BIT = Bits(bit_length=1)
BIT_N = Bits(bit_length=1, negated=True)
STR = HString()
SLICE = HSlice()
FLOAT64 = HFloat(11, 52, name="float64")




from copy import deepcopy
from functools import reduce
from itertools import compress
from operator import and_
from typing import List, Tuple, Dict, Optional, Callable, Set

from hwt.doc_markers import internal
from hwt.hdl.operatorUtils import replace_input_in_expr
from hwt.hdl.sensitivityCtx import SensitivityCtx
from hwt.hdl.statements.statement import HdlStatement
from hwt.hdl.statements.utils.comparison import  statementsAreSame, isSameStatementList
from hwt.hdl.statements.utils.ioDiscovery import HdlStatement_discover_enclosure_for_statements
from hwt.hdl.statements.utils.reduction import HdlStatement_merge_statement_lists, \
    HdlStatement_try_reduce_list, is_mergable_statement_list
from hwt.hdl.statements.utils.signalCut import HdlStatement_cut_off_drivers_of_list
from hwt.serializer.utils import RtlSignal_sort_key
from hwt.synthesizer.rtlLevel.fill_stm_list_with_enclosure import fill_stm_list_with_enclosure
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.walkers import discover_sensitivity_of_sig


class IfContainer(HdlStatement):
    """
    Structural container of if statement for hdl rendering

    :ivar ~._ifTrue_enclosed_for: set of signals for which if ifTrue branch enclosed
            (has not branch where signal is not assignment)
    :ivar ~._elIfs_enclosed_for: list of sets of enclosed signals for each elif
    :ivar ~._ifFalse_enclosed_for: set of enclosed signals for ifFalse branch
    """
    _DEEPCOPY_SHALLOW_ONLY = (*HdlStatement._DEEPCOPY_SHALLOW_ONLY, '_ifTrue_enclosed_for', '_elIfs_enclosed_for', '_ifFalse_enclosed_for')
    _DEEPCOPY_SKIP = (*HdlStatement._DEEPCOPY_SKIP, 'cond', 'elIfs')

    def __init__(self, cond: RtlSignalBase, ifTrue=None, ifFalse=None, elIfs=None,
                 parentStm=None, event_dependent_from_branch: Optional[int]=None):
        """
        :param cond: RtlSignal as conditions for this if
        :param ifTrue: list of statements which should be active if cond.
            is met
        :param elIfs: list of tuples (list of conditions, list of statements)
        :param ifFalse: list of statements which should be active if cond.
            and any other cond. in elIfs is met
        """
        assert isinstance(cond, RtlSignalBase)
        self.cond = cond
        super(IfContainer, self).__init__(
            parentStm,
            event_dependent_from_branch=event_dependent_from_branch)

        if ifTrue is None:
            ifTrue = []
        self.ifTrue: List[HdlStatement] = ifTrue

        if elIfs is None:
            elIfs = []
        self.elIfs: List[Tuple[RtlSignalBase, List[HdlStatement]]] = elIfs

        self.ifFalse: Optional[List[HdlStatement]] = ifFalse
        self._ifTrue_enclosed_for: Optional[Set[RtlSignalBase]] = None
        self._elIfs_enclosed_for: Optional[Set[RtlSignalBase]] = None
        self._ifFalse_enclosed_for: Optional[Set[RtlSignalBase]] = None

    def __deepcopy__(self, memo: dict):
        result = super(IfContainer, self).__deepcopy__(memo)
        result.cond = self.cond
        result.elIfs = [(c, deepcopy(stms, memo)) for c, stms in self.elIfs]
        return result

    @internal
    def _collect_io(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._collect_io`
        """
        if isinstance(self.cond, RtlSignalBase):
            self._inputs.append(self.cond)
        for c, _ in self.elIfs:
            if isinstance(c, RtlSignalBase):
                self._inputs.append(c)
        super(IfContainer, self)._collect_io()

    @internal
    def _collect_inputs(self) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._collect_inputs`
        """
        if isinstance(self.cond, RtlSignalBase):
            self._inputs.append(self.cond)
        for c, _ in self.elIfs:
            if isinstance(c, RtlSignalBase):
                self._inputs.append(c)
        super(IfContainer, self)._collect_inputs()

    @internal
    def _clean_signal_meta(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._clean_signal_meta`
        """
        self._sensitivity = None
        self._ifTrue_enclosed_for = None
        self._elIfs_enclosed_for = None
        self._ifFalse_enclosed_for = None
        HdlStatement._clean_signal_meta(self)

    @internal
    def _cut_off_drivers_of(self, sig: RtlSignalBase):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._cut_off_drivers_of`
        """
        if self._sensitivity is not None or self._enclosed_for is not None:
            raise NotImplementedError(
                    "Sensitivity and enclosure has to be cleaned first")

        if len(self._outputs) == 1 and sig in self._outputs:
            # this statement has only this output, eject this statement from its parent
            self.parentStm = None  # because new parent will be asigned immediately after cutting of
            return self

        sig.drivers.discard(self)
        # try to cut off all statements which are drivers of specified signal
        # in all branches
        child_keep_mask = []

        newIfTrue = []
        all_cut_off = True
        all_cut_off &= HdlStatement_cut_off_drivers_of_list(
            sig, self.ifTrue, child_keep_mask, newIfTrue)
        self.ifTrue = list(compress(self.ifTrue, child_keep_mask))

        newElifs = []
        anyElifHit = False
        for cond, stms in self.elIfs:
            newCase = []
            child_keep_mask.clear()
            all_cut_off &= HdlStatement_cut_off_drivers_of_list(
                sig, stms, child_keep_mask, newCase)

            _stms = list(compress(stms, child_keep_mask))
            stms.clear()
            stms.extend(_stms)

            if newCase:
                anyElifHit = True
            newElifs.append((cond, newCase))

        newIfFalse = None
        if self.ifFalse:
            newIfFalse = []
            child_keep_mask.clear()
            all_cut_off &= HdlStatement_cut_off_drivers_of_list(
                sig, self.ifFalse, child_keep_mask, newIfFalse)
            self.ifFalse = list(compress(self.ifFalse, child_keep_mask))

        assert not all_cut_off, "everything was cut of but this should be already known at the start"

        if newIfTrue or newIfFalse or anyElifHit or newIfFalse:
            # parts were cut off
            # generate new statement for them
            cond_sig = self.cond
            n = self.__class__(cond_sig, newIfTrue)
            for c_sig, stms in newElifs:
                n.Elif(c_sig, stms)
            if newIfFalse is not None:
                n.Else(newIfFalse)

            if self.parentStm is None:
                ctx = n._get_rtl_context()
                ctx.statements.add(n)

            self._cut_off_drivers_of_regenerate_io(sig, n)

            return n

    @internal
    def _discover_enclosure(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._discover_enclosure`
        """
        outputs = self._outputs
        self._ifTrue_enclosed_for = HdlStatement_discover_enclosure_for_statements(
            self.ifTrue, outputs)

        elif_encls = self._elIfs_enclosed_for = []
        for _, stms in self.elIfs:
            e = HdlStatement_discover_enclosure_for_statements(
                stms, outputs)
            elif_encls.append(e)

        self._ifFalse_enclosed_for = HdlStatement_discover_enclosure_for_statements(
            self.ifFalse, outputs)

        assert self._enclosed_for is None
        encl = self._enclosed_for = set()

        for s in self._ifTrue_enclosed_for:
            enclosed = True

            for elif_e in elif_encls:
                if s not in elif_e:
                    enclosed = False
                    break

            if enclosed and s in self._ifFalse_enclosed_for:
                encl.add(s)

    @internal
    def _discover_sensitivity(self, seen: set) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._discover_sensitivity`
        """
        assert self._sensitivity is None, self
        ctx = self._sensitivity = SensitivityCtx()

        discover_sensitivity_of_sig(self.cond, seen, ctx)
        if ctx.contains_ev_dependency:
            return

        for stm in self.ifTrue:
            stm._discover_sensitivity(seen)
            ctx.extend(stm._sensitivity)

        # elifs
        for cond, stms in self.elIfs:
            if ctx.contains_ev_dependency:
                break

            discover_sensitivity_of_sig(cond, seen, ctx)
            if ctx.contains_ev_dependency:
                break

            for stm in stms:
                if ctx.contains_ev_dependency:
                    break

                stm._discover_sensitivity(seen)
                ctx.extend(stm._sensitivity)

        if self.ifFalse:
            assert not ctx.contains_ev_dependency, "can not negate event"
            # else
            for stm in self.ifFalse:
                stm._discover_sensitivity(seen)
                ctx.extend(stm._sensitivity)

    @internal
    def _fill_enclosure(self, enclosure: Dict[RtlSignalBase, Callable[[], HdlStatement]]) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._fill_enclosure`
        """
        enc = []
        outputs = self._outputs
        for e in sorted(enclosure.keys(), key=RtlSignal_sort_key):
            if e in outputs and e not in self._enclosed_for:
                enc.append(e)

        if not enc:
            return
        fill_stm_list_with_enclosure(self, self._ifTrue_enclosed_for,
                                     self.ifTrue, enc, enclosure)

        for (_, stms), e in zip(self.elIfs, self._elIfs_enclosed_for):
            fill_stm_list_with_enclosure(self, e, stms, enc, enclosure)

        self.ifFalse = fill_stm_list_with_enclosure(self, self._ifFalse_enclosed_for,
                                                    self.ifFalse, enc, enclosure)

        self._enclosed_for.update(enc)

    def _iter_stms(self):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._iter_stms`
        """
        yield from self.ifTrue
        for _, stms in self.elIfs:
            yield from stms
        if self.ifFalse is not None:
            yield from self.ifFalse

    @internal
    def _try_reduce(self) -> Tuple[bool, List[HdlStatement]]:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._try_reduce`
        """
        # flag if IO of statement has changed
        io_change = False

        self.ifTrue, rank_decrease, _io_change = HdlStatement_try_reduce_list(
            self.ifTrue)
        self.rank -= rank_decrease
        io_change |= _io_change

        new_elifs = []
        for cond, statements in self.elIfs:
            _statements, rank_decrease, _io_change = HdlStatement_try_reduce_list(
                statements)
            self.rank -= rank_decrease
            io_change |= _io_change
            new_elifs.append((cond, _statements))
        self.elIfs = new_elifs

        if self.ifFalse is not None:
            self.ifFalse, rank_decrease, _io_update_required = HdlStatement_try_reduce_list(
                self.ifFalse)
            self.rank -= rank_decrease
            io_change |= _io_change

        reduce_self = not self._condHasEffect(
            self.ifTrue, self.ifFalse, self.elIfs)

        if reduce_self:
            res = self.ifTrue
        else:
            res = [self, ]

        self._on_reduce(reduce_self, io_change, res)

        # try merge nested ifs as elifs
        if self.ifFalse is not None and len(self.ifFalse) == 1:
            child = self.ifFalse[0]
            if isinstance(child, IfContainer):
                self._merge_nested_if_from_else(child)

        return res, io_change

    @internal
    def _merge_nested_if_from_else(self, ifStm: "IfContainer"):
        """
        Merge nested IfContarner form else branch to this IfContainer
        as elif and else branches
        """
        self.elIfs.append((ifStm.cond, ifStm.ifTrue))
        self.elIfs.extend(ifStm.elIfs)

        self.ifFalse = ifStm.ifFalse

    @internal
    def _is_mergable(self, other: HdlStatement) -> bool:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._is_mergable`
        """
        if not isinstance(other, IfContainer):
            return False

        if (self.cond is not other.cond
                or not is_mergable_statement_list(self.ifTrue, other.ifTrue)):
            return False

        if len(self.elIfs) != len(other.elIfs):
            return False

        for (a_c, a_stm), (b_c, b_stm) in zip(self.elIfs, other.elIfs):
            if a_c is not b_c or not is_mergable_statement_list(a_stm, b_stm):
                return False

        if not is_mergable_statement_list(self.ifFalse, other.ifFalse):
            return False
        return True

    @internal
    def _merge_with_other_stm(self, other: "IfContainer") -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._merge_with_other_stm`
        """
        merge = HdlStatement_merge_statement_lists
        self.ifTrue = merge(self.ifTrue, other.ifTrue)

        new_elifs = []
        for ((c, elifA), (_, elifB)) in zip(self.elIfs, other.elIfs):
            new_elifs.append((c, merge(elifA, elifB)))
        self.elIfs = new_elifs

        self.ifFalse = merge(self.ifFalse, other.ifFalse)

        other.ifTrue = []
        other.elIfs = []
        other.ifFalse = None

        self._on_merge(other)

    @internal
    @staticmethod
    def _condHasEffect(ifTrue, ifFalse, elIfs):
        stmCnt = len(ifTrue)
        # [TODO] condition in empty if stm
        if ifFalse is not None \
                and stmCnt == len(ifFalse) \
                and reduce(and_,
                           [len(stm) == stmCnt
                            for _, stm in elIfs],
                           True):
            for stms in zip(ifTrue, ifFalse, *map(lambda x: x[1], elIfs)):
                if not statementsAreSame(stms):
                    return True
            return False
        return True

    def isSame(self, other: HdlStatement) -> bool:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement.isSame`
        """
        if self is other:
            return True

        if self.rank != other.rank:
            return False

        if isinstance(other, IfContainer):
            if self.cond is other.cond:
                if len(self.ifTrue) == len(other.ifTrue) \
                        and ((self.ifFalse is None and other.ifFalse is None) or
                             len(self.ifFalse) == len(other.ifFalse)) \
                        and len(self.elIfs) == len(other.elIfs):
                    if not isSameStatementList(self.ifTrue,
                                               other.ifTrue) \
                            or not isSameStatementList(self.ifFalse,
                                                       other.ifFalse):
                        return False
                    for (ac, astms), (bc, bstms) in zip(self.elIfs,
                                                        other.elIfs):
                        if not (ac == bc) or\
                                not isSameStatementList(astms, bstms):
                            return False
                    return True
        return False

    @internal
    def _replace_input(self, toReplace: RtlSignalBase, replacement: RtlSignalBase):
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._replace_input`
        """
        isTopStm = self.parentStm is None
        if isTopStm:
            self.cond = replace_input_in_expr(self, self.cond, toReplace,
                                              replacement, isTopStm)

        for stm in self.ifTrue:
            stm._replace_input(toReplace, replacement)

        new_elifs = []
        for (cond, stms) in self.elIfs:
            new_cond = replace_input_in_expr(self, cond, toReplace, replacement, isTopStm)
            for stm in stms:
                stm._replace_input(toReplace, replacement)
            new_elifs.append((new_cond, stms))
        self.elIfs = new_elifs

        if self.ifFalse is not None:
            for stm in self.ifFalse:
                stm._replace_input(toReplace, replacement)

        self._replace_input_update_sensitivity_and_enclosure(toReplace, replacement)

    @internal
    def _replace_child_statement(self, stm:HdlStatement,
            replacement:List[HdlStatement],
            update_io:bool) -> None:
        """
        :see: :meth:`hwt.hdl.statements.statement.HdlStatement._replace_child_statement`
        """

        if update_io:
            raise NotImplementedError()
        for branch_list in (self.ifTrue, *(elif_stms for _, elif_stms in self.elIfs), self.ifFalse):
            if branch_list is None:
                continue
            try:
                i = branch_list.index(stm)
            except ValueError:
                # not in list
                continue

            self.rank -= stm.rank
            branch_list[i:i + 1] = replacement
            for rstm in replacement:
                rstm._set_parent_stm(self)
            # reset IO because it was shared with this statement
            stm._destroy()
            return

        raise ValueError("Statement", stm, "not found in ", self)




from copy import copy
from operator import lshift, rshift

from hwt.doc_markers import internal
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements.statement import HwtSyntaxError
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.sliceUtils import slice_to_SLICE
from hwt.hdl.types.typeCast import toHVal
from hwt.synthesizer.exceptions import TypeConversionErr
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import SignalDriverErr
from hwt.synthesizer.vectorUtils import fitTo_t


def tv(signal):
    """
    HValue class for hdl type of signal
    """
    return signal._dtype.getValueCls()


class RtlSignalOps():
    """
    Definitions of operators and other operator functions for RtlSignal

    :ivar ~._usedOps: cache for expressions with this signal
    """

    def _auto_cast(self, toT):
        try:
            return self._dtype.auto_cast(self, toT)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _reinterpret_cast(self, toT):
        try:
            return self._dtype.reinterpret_cast(self, toT)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    @internal
    def naryOp(self, operator, indexOfSelfInOperands, opCreateDelegate , *otherOps) -> RtlSignalBase:
        """
        Try lookup operator with this parameters in _usedOps
        if not found create new one and stored it in _usedOps

        :param operator: instance of OpDefinition
        :param opCreateDelegate: function (\*ops) to create operator
        :param otherOps: other operands (ops = self + otherOps)

        :return: RtlSignal which is result of newly created operator
        """
        k = (operator, indexOfSelfInOperands, *otherOps)
        used = self._usedOps
        try:
            return used[k]
        except KeyError:
            pass

        if indexOfSelfInOperands == 0:
            o = opCreateDelegate(self, *otherOps)
        else:
            o = opCreateDelegate(*otherOps[:indexOfSelfInOperands], self, *otherOps[indexOfSelfInOperands])
        # input operads may be type converted,
        # search if this happend, and return always same result signal
        try:
            op_instantiated = (o.origin.operator == operator
                               and o.origin.operands[indexOfSelfInOperands] is self)
        except AttributeError:
            op_instantiated = False

        usedOpsAlias = self._usedOpsAlias
        if op_instantiated:
            # try check real operands and operator which were used after all default type conversions
            k_real = (operator, indexOfSelfInOperands, *o.origin.operands[1:])
            if k != k_real:
                alias = usedOpsAlias[k_real]
                usedOpsAlias[k] = alias
                alias.add(k)
                used[k] = o

        return o

    def __invert__(self):
        try:
            return self.naryOp(AllOps.NOT, 0, tv(self).__invert__)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _onRisingEdge(self):
        try:
            return self.naryOp(AllOps.RISING_EDGE, 0, tv(self)._onRisingEdge)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _onFallingEdge(self):
        try:
            return self.naryOp(AllOps.FALLING_EDGE, 0, tv(self)._onFallingEdge)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _isOn(self):
        try:
            return self._auto_cast(BOOL)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    # conversions
    def _convSign(self, signed):
        try:
            return tv(self)._convSign(self, signed)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _signed(self):
        try:
            return tv(self)._signed(self)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _unsigned(self):
        try:
            return tv(self)._unsigned(self)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _vec(self):
        try:
            return tv(self)._vec(self)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    # logic
    def __and__(self, other):
        try:
            return self.naryOp(AllOps.AND, 0, tv(self).__and__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __xor__(self, other):
        try:
            return self.naryOp(AllOps.XOR, 0, tv(self).__xor__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __or__(self, other):
        try:
            return self.naryOp(AllOps.OR, 0, tv(self).__or__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __lshift__(self, other):
        try:
            return self.naryOp(lshift, 0, tv(self).__lshift__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __rshift__(self, other):
        try:
            return self.naryOp(rshift, 0, tv(self).__rshift__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    # cmp
    def _eq(self, other):
        """
        __eq__ is not overloaded because it will destroy hashability of object
        """
        try:
            return self.naryOp(AllOps.EQ, 0, tv(self)._eq, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __ne__(self, other):
        try:
            return self.naryOp(AllOps.NE, 0, tv(self).__ne__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __ge__(self, other):
        try:
            return self.naryOp(AllOps.GE, 0, tv(self).__ge__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __gt__(self, other):
        try:
            return self.naryOp(AllOps.GT, 0, tv(self).__gt__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __lt__(self, other):
        try:
            return self.naryOp(AllOps.LT, 0, tv(self).__lt__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __le__(self, other):
        try:
            return self.naryOp(AllOps.LE, 0, tv(self).__le__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    # arithmetic
    def __neg__(self):
        try:
            return self.naryOp(AllOps.ADD, 0, tv(self).__neg__)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __add__(self, other):
        try:
            return self.naryOp(AllOps.ADD, 0, tv(self).__add__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __sub__(self, other):
        try:
            return self.naryOp(AllOps.SUB, 0, tv(self).__sub__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __mul__(self, other):
        try:
            return self.naryOp(AllOps.MUL, 0, tv(self).__mul__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __mod__(self, other):
        try:
            return self.naryOp(AllOps.MOD, 0, tv(self).__mod__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __pow__(self, other):
        try:
            return self.naryOp(AllOps.POW, 0, tv(self).__pow__, other)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __floordiv__(self, divider):
        try:
            return self.naryOp(AllOps.DIV, 0, tv(self).__floordiv__, divider)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __getitem__(self, key):
        try:
            if isinstance(key, slice):
                key = slice_to_SLICE(key, self._dtype.bit_length())
            return self.naryOp(AllOps.INDEX, 0, tv(self).__getitem__, key)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _concat(self, *operands):
        try:
            return self.naryOp(AllOps.CONCAT, 0, tv(self)._concat, *operands)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def _ternary(self, ifTrue, ifFalse):
        try:
            return self.naryOp(AllOps.TERNARY, 0, tv(self)._ternary, ifTrue, ifFalse)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    @internal
    def _getIndexCascade(self):
        """
        Find out if this signal is something indexed
        """
        intf = self
        indexes = []
        while True:
            try:
                # now I am result of the index  xxx[xx] <= source
                # get index op
                d = intf.singleDriver()
                try:
                    op = d.operator
                except AttributeError:
                    break

                if op == AllOps.INDEX:
                    # get signal on which is index applied
                    indexedOn = d.operands[0]
                    if isinstance(indexedOn, RtlSignalBase):
                        intf = indexedOn
                        indexes.append(d.operands[1])
                    else:
                        raise HwtSyntaxError(
                            f"can not assign to a static value {indexedOn}")
            except SignalDriverErr:
                break

        if not indexes:
            indexes = None
        else:
            indexes.reverse()

        return intf, indexes

    def _getDestinationSignalForAssignmentToThis(self):
        """
        :return: a signal which should be used as a destination if assigning to this signal
        """
        return self

    def __call__(self, source,
                 dst_resolve_fn=lambda x: x._getDestinationSignalForAssignmentToThis(),
                 exclude=None,
                 fit=False) -> HdlAssignmentContainer:
        """
        Create assignment to this signal

        :attention: it is not call of function it is operator of assignment
        :return: list of assignments
        """
        assert not self._const, self
        if exclude is not None and (self in exclude or source in exclude):
            return []

        if self.hidden:
            try:
                d = self.singleDriver()
            except:
                d = None
            operator = getattr(d, "operator", None)
            if operator is not None:
                if operator.allowsAssignTo:
                    pass
                elif operator == AllOps.NOT:
                    return d.operands[0](~source, dst_resolve_fn=dst_resolve_fn, exclude=exclude, fit=fit)
                else:
                    raise AssertionError("Assignment to", self, "is not allowed by operator definition")

        if isinstance(source, InterfaceBase):
            assert source._isAccessible, (source, "must be a Signal Interface which is accessible in current scope")
            source = source._sig

        try:
            if source is None:
                requires_type_check = False
                source = self._dtype.from_py(None)
            else:
                requires_type_check = True
                source = toHVal(source, suggestedType=self._dtype)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

        if requires_type_check:
            err = False
            try:
                if fit:
                    source = fitTo_t(source, self._dtype)
                source = source._auto_cast(self._dtype)
            except TypeConversionErr:
                err = True
            if err:
                raise TypeConversionErr(
                    ("Can not connect %r (of type %r) to %r "
                     "(of type %r) due type incompatibility")
                    % (source, source._dtype, self, self._dtype))
        try:
            mainSig, indexCascade = self._getIndexCascade()
            mainSig = dst_resolve_fn(mainSig)
            return HdlAssignmentContainer(source, mainSig, indexCascade)
        except Exception as e:
            # simplification of previous exception traceback
            e_simplified = copy(e)
            raise e_simplified

    def __int__(self):
        if not self._const:
            raise TypeError("Int value of signal can not be evaluated"
                            " because it is not constant expression:", self)
        else:
            return int(self._val)

    def __bool__(self):
        if not self._const:
            raise TypeError("Bool value of signal can be evaluated"
                            " because it is not constant expression:", self)
        else:
            return bool(self._val)

    def _is_full_valid(self):
        return self._const and self._val._is_full_valid()



"""
Sim model serializer serialize HDL objects to simulation model writen in python.
For hwtSimApi.basic_rtl_simulator
"""
from hdlConvertorAst.to.basic_hdl_sim_model._main import ToBasicHdlSimModel
from hwt.serializer.serializer_config import DummySerializerConfig
from hwt.serializer.simModel.serializer import ToHdlAstSimModel


class SimModelSerializer(DummySerializerConfig):
    fileExtension = '.py'
    TO_HDL_AST = ToHdlAstSimModel
    TO_HDL = ToBasicHdlSimModel



from copy import copy

from hwt.doc_markers import internal
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import INT
from hwt.hdl.value import HValue
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


def slice_member_to_hval(v):
    if isinstance(v, RtlSignalBase):  # is signal
        assert isinstance(v._dtype, Bits)
        return v
    elif isinstance(v, HValue):
        if isinstance(v, Bits):
            return v
        else:
            return v._auto_cast(INT)
    else:
        return INT.from_py(v)


class HSliceVal(HValue):
    """
    HValue class for HSlice type
    """

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        assert vld_mask is None, vld_mask
        if val is None:
            val = slice(None, None, None)
        else:
            assert isinstance(val, slice), val
            start = slice_member_to_hval(val.start)
            stop = slice_member_to_hval(val.stop)
            step = slice_member_to_hval(val.step)
            val = slice(start, stop, step)

        return cls(typeObj, val, vld_mask=1)

    def _is_full_valid(self):
        v = self.val
        return v.start._is_full_valid() and v.stop._is_full_valid()

    def to_py(self):
        """
        Convert to python slice object
        """
        v = self.val
        return slice(int(v.start), int(v.stop), int(v.step))

    def _size(self):
        """
        :return: how many bits is this slice selecting
        """
        assert isinstance(self, HValue)
        v = self.val
        if v.step == -1:
            return int(v.start) - int(v.stop)
        elif v.step == 1:
            return int(v.stop) - int(v.start)
        else:
            raise NotImplementedError(self)

    def _eq_val(self, other):
        assert isinstance(other, HSliceVal)
        return self.val == other.val

    def _eq(self, other):
        return self._eq__val(other)

    def __lt__(self, other):
        if self.val.step != other.val.step:
            raise ValueError()
        if isinstance(other, INT.getValueCls()):
            return self.val.start < other
        else:
            return (self.val.start, self.val.stop) < (other.val.start, other.val.stop)

    def __copy__(self):
        v = HValue.__copy__(self)
        v.val = copy(v.val)
        return v

    def staticEval(self):
        v = self.val
        new_v = slice(
            v.start.staticEval(),
            v.stop.staticEval(),
            v.step.staticEval(),
        )
        return self.__class__.from_py(self._dtype, new_v)

    @internal
    def __hash__(self):
        v = self.val
        return hash((self._dtype, v.start, v.stop, v.step))



from itertools import chain
from typing import Union, Optional, Tuple

from hwt.doc_markers import internal
from hwt.hdl.operatorDefs import OpDefinition
from hwt.hdl.types.array import HArray
from hwt.hdl.types.defs import BIT
from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.struct import HStruct
from hwt.interfaces.std import Signal, Clk, Rst, Rst_n
from hwt.interfaces.structIntf import HdlType_to_Interface, StructIntf
from hwt.synthesizer.hObjList import HObjList
from hwt.synthesizer.interfaceLevel.getDefaultClkRts import getClk, getRst
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase, InterfaceBase
from hwt.synthesizer.rtlLevel.constants import NOT_SPECIFIED
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.synthesizer.rtlLevel.rtlSyncSignal import RtlSyncSignal
from ipCorePackager.constants import INTF_DIRECTION


def getSignalName(sig):
    """
    Name getter which works for RtlSignal and Interface instances as well
    """
    try:
        return sig._name
    except AttributeError:
        pass
    return sig.name


@internal
def _default_param_updater(self, myP, otherP_val):
    myP.set_value(otherP_val)


@internal
def _normalize_default_value_dict_for_interface_array(root_val: dict,
                                                      val: Union[dict, list, None],
                                                      name_prefix: str,
                                                      hobj_list: HObjList,
                                                      neutral_value):
    """
    This function is called to convert data in format
    .. code-block:: python

        {"x": [3, 4]}
        # into
        {"x_0": 3, "x_1": 4}

    This is required because the items of HObjList are stored in _interfaces as a separate items
    and thus we can not resolve the value association otherwise.
    """

    for i, intf in enumerate(hobj_list):
        if val is neutral_value:
            continue
        elif isinstance(val, dict):
            _val = val.get(i, neutral_value)
        else:
            _val = val[i]
        if _val is neutral_value:
            continue

        elm_name = f"{name_prefix:s}_{i:d}"
        if isinstance(intf, HObjList):
            _normalize_default_value_dict_for_interface_array(root_val, _val, elm_name, intf, neutral_value)
        else:
            root_val[elm_name] = _val


@internal
def _instanciate_signals(intf: Union[Signal, HObjList,StructIntf], clk: Clk, rst: Union[Rst, Rst_n], def_val, nop_val, signal_create_fn):
    intf._direction = INTF_DIRECTION.UNKNOWN
    if isinstance(intf, Signal):
        name = intf._getHdlName()
        intf._sig = signal_create_fn(
            name,
            intf._dtype,
            clk, rst, def_val, nop_val)

    elif isinstance(intf, HObjList):
        intf_len = len(intf)
        if isinstance(def_val, dict):
            for k in def_val.keys():
                assert k > 0 and k < intf_len, ("Default value for", intf, " specifies ", k, " which is not present on interface")
        elif def_val is not None:
            assert len(def_val) == intf_len, ("Default value does not have same size, ", len(def_val), intf_len, intf)

        if isinstance(nop_val, dict):
            for k in nop_val.keys():
                assert k > 0 and k < intf_len, ("Nop value for", intf, " specifies ", k, " which is not present on interface")
        elif nop_val is not NOT_SPECIFIED:
            assert len(nop_val) == intf_len, ("Nop value does not have same size, ", len(nop_val), intf_len, intf)

        for i, elm in enumerate(intf):
            if def_val is None:
                _def_val = None
            elif isinstance(def_val, dict):
                _def_val = def_val.get(i, None)
            else:
                _def_val = def_val[i]

            if nop_val is NOT_SPECIFIED:
                _nop_val = NOT_SPECIFIED
            elif isinstance(nop_val, dict):
                _nop_val = nop_val.get(i, NOT_SPECIFIED)
            else:
                _nop_val = nop_val[i]
            _instanciate_signals(elm, clk, rst, _def_val, _nop_val, signal_create_fn)

    elif isinstance(intf, StructIntf):
        if def_val is not None:
            for k in tuple(def_val.keys()):
                _i = getattr(intf, k, NOT_SPECIFIED)
                assert _i is not NOT_SPECIFIED, ("Default value for", intf, " specifies ", k, " which is not present on interface")
                if isinstance(_i, HObjList):
                    _normalize_default_value_dict_for_interface_array(
                        def_val, def_val[k], k, _i, None)

        if nop_val is not NOT_SPECIFIED:
            for k in tuple(nop_val.keys()):
                _i = getattr(intf, k, NOT_SPECIFIED)
                assert _i is not NOT_SPECIFIED, ("Nop value for", intf, " specifies ", k, " which is not present on interface")
                if isinstance(_i, HObjList):
                    _normalize_default_value_dict_for_interface_array(
                        nop_val, nop_val[k],
                        k, _i, NOT_SPECIFIED)

        for elm in intf._interfaces:
            name = elm._name
            if def_val is None:
                _def_val = None
            else:
                _def_val = def_val.get(name, None)

            if nop_val is NOT_SPECIFIED:
                _nop_val = NOT_SPECIFIED
            else:
                _nop_val = nop_val.get(name, NOT_SPECIFIED)

            _instanciate_signals(elm, clk, rst, _def_val, _nop_val, signal_create_fn)
    else:
        raise NotImplementedError(intf)


@internal
def _loadDeclarations(intf_or_list: Union[HObjList, InterfaceBase], name: str):
    if isinstance(intf_or_list, HObjList):
        for i, intf in enumerate(intf_or_list):
            _loadDeclarations(intf, f"{name:s}_{i:d}")
    else:
        intf_or_list._name = name
        intf_or_list._loadDeclarations()


class UnitImplHelpers(UnitBase):

    def _reg(self, name: str,
             dtype: HdlType=BIT,
             def_val: Union[int, None, dict, list]=None,
             clk: Union[RtlSignalBase, None, Tuple[RtlSignalBase, OpDefinition]]=None,
             rst: Optional[RtlSignalBase]=None) -> RtlSyncSignal:
        """
        Create RTL FF register in this unit

        :param def_val: s default value of this register,
            if this value is specified reset signal of this component is used
            to generate a reset logic
        :param clk: optional clock signal specification,
            (signal or tuple(signal, edge type (AllOps.RISING_EDGE/FALLING_EDGE)))
        :param rst: optional reset signal specification
        :note: rst/rst_n resolution is done from signal type,
            if it is negated type the reset signal is interpreted as rst_n
        :note: if clk or rst is not specified default signal
            from parent unit instance will be used
        """
        if clk is None:
            clk = getClk(self)

        if def_val is None:
            # if no value is specified reset is not required
            rst = None
        elif rst is None:
            rst = getRst(self)

        if isinstance(dtype, (HStruct, HArray)):
            container = HdlType_to_Interface().apply(dtype)
            _loadDeclarations(container, name)
            _instanciate_signals(
                container, clk, rst, def_val, NOT_SPECIFIED,
                lambda name, dtype, clk, rst, def_val, nop_val: self._reg(name, dtype,
                                                                          def_val=def_val,
                                                                          clk=clk, rst=rst))
            container._parent = self
            return container
        else:
            # primitive data type signal
            return self._ctx.sig(
                name,
                dtype=dtype,
                clk=clk,
                syncRst=rst,
                def_val=def_val
            )

    def _sig(self, name: str,
             dtype: HdlType=BIT,
             def_val: Union[int, None, dict, list]=None,
             nop_val: Union[int, None, dict, list, "NOT_SPECIFIED"]=NOT_SPECIFIED) -> RtlSignal:
        """
        Create signal in this unit

        :see: :func:`hwt.synthesizer.rtlLevel.netlist.RtlNetlist.sig`
        """
        if isinstance(dtype, HStruct):
            container = HdlType_to_Interface().apply(dtype)
            _loadDeclarations(container, name)
            _instanciate_signals(
                container, None, None, def_val, nop_val,
                lambda name, dtype, clk, rst, def_val, nop_val: self._sig(name, dtype,
                                                                          def_val=def_val,
                                                                          nop_val=nop_val))
            container._parent = self
            return container
        else:
            # primitive data type signal
            return self._ctx.sig(name, dtype=dtype, def_val=def_val, nop_val=nop_val)

    @internal
    def _cleanAsSubunit(self):
        """
        Disconnect internal signals so unit can be reused by parent unit
        """
        for i in chain(self._interfaces, self._private_interfaces):
            i._clean()

    @internal
    def _signalsForSubUnitEntity(self, context, prefix: str):
        """
        generate signals in this context for all ports of this subunit
        """
        for i in self._interfaces:
            if i._isExtern:
                i._signalsForInterface(context, None, None, prefix=prefix + i._NAME_SEPARATOR)




from hdlConvertorAst.hdlAst._expr import HdlTypeAuto, HdlValueId, HdlOp, \
    HdlOpType
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_index, \
    hdl_downto
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import INT, FLOAT64
from hwt.hdl.types.hdlType import HdlType, MethodNotOverloaded
from hwt.serializer.verilog.utils import SIGNAL_TYPE
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal
from hwt.hdl.types.float import HFloat


class ToHdlAstVerilog_types():
    INT = HdlValueId("int", obj=int)
    REG = HdlValueId("reg", obj=LanguageKeyword())
    WIRE = HdlValueId("wire", obj=LanguageKeyword())

    def does_type_requires_extra_def(self, t: HdlType, other_types: list):
        try:
            return t._as_hdl_requires_def(self, other_types)
        except MethodNotOverloaded:
            pass
        return False

    def as_hdl_HdlType_bits(self, typ: Bits, declaration=False):
        isVector = typ.force_vector or typ.bit_length() > 1
        sigType = self.signalType

        if typ == INT:
            t = self.INT
        elif sigType is SIGNAL_TYPE.PORT_WIRE:
            t = HdlTypeAuto
        elif sigType is SIGNAL_TYPE.REG or sigType is SIGNAL_TYPE.PORT_REG:
            t = self.REG
        elif sigType is SIGNAL_TYPE.WIRE:
            t = self.WIRE
        else:
            raise ValueError(sigType)

        if typ.signed is None or typ.signed == False:
            # [yosys] do not produce unsigned type as yosys support is limited
            is_signed = None
        else:
            is_signed = self.as_hdl_int(int(typ.signed))

        if isVector:
            w = typ.bit_length()
            assert isinstance(w, int) or (isinstance(w, RtlSignal) and w._const), w
            w = hdl_downto(self.as_hdl(w - 1),
                           self.as_hdl_int(0))
        else:
            w = None

        return HdlOp(HdlOpType.PARAMETRIZATION, [t, w, is_signed])

    def as_hdl_HdlType_array(self, typ: HArray, declaration=False):
        if declaration:
            raise NotImplementedError()
        else:
            _int = self.as_hdl_int
            size = HdlOp(HdlOpType.DOWNTO, [_int(0),
                                                 _int(int(typ.size) - 1)])
            return hdl_index(self.as_hdl_HdlType(typ.element_t), size)

    def as_hdl_HdlType_enum(self, typ, declaration=False):
        if declaration:
            raise TypeError(
                "Target language does not use enum types, this library should uses Bits instead"
                " (this should not be required because it should have been filtered before)")
        else:
            valueCnt = len(typ._allValues)
            return self.as_hdl_HdlType_bits(Bits(valueCnt.bit_length()),
                                            declaration=declaration)

    def as_hdl_HdlType_float(self, typ: HFloat, declaration=False):
        if typ == FLOAT64:
            return HdlValueId("real")
        else:
            raise NotImplementedError(typ)



from hwt.synthesizer.rtlLevel.extract_part_drivers import extract_part_drivers
from hwt.synthesizer.rtlLevel.mark_visibility_of_signals_and_check_drivers import markVisibilityOfSignalsAndCheckDrivers
from hwt.synthesizer.rtlLevel.remove_unconnected_signals import removeUnconnectedSignals


class DummyPlatform():
    """
    Dummy synthesis platform, base class of all chip and toolset specific platforms.
    Plaform in this context is a set of configurations which do describe the target toolset and chip/node.
    It can also contains pre/post processing callbacks and optimizations required for this target.

    :note: all processors has to be callable with only one parameter
        which is actual Unit/RtlNetlist instance
    """

    def __init__(self):
        self.beforeToRtl = []
        self.beforeToRtlImpl = []
        self.afterToRtlImpl = []

        self.beforeHdlArchGeneration = [
            extract_part_drivers,
            removeUnconnectedSignals,
            markVisibilityOfSignalsAndCheckDrivers,
        ]
        self.afterToRtl = []



from hwt.doc_markers import internal
from hwt.hdl.types.defs import INT, SLICE
from hwt.hdl.types.slice import HSlice
from hwt.hdl.types.typeCast import toHVal


@internal
def slice_to_SLICE(sliceVals, width):
    """convert python slice to value of SLICE hdl type"""
    if sliceVals.step is None:
        step = -1
    else:
        step = sliceVals.step

    start = sliceVals.start
    if start is None:
        start = INT.from_py(width)
    else:
        start = toHVal(start)

    stop = sliceVals.stop
    if stop is None:
        stop = INT.from_py(0)
    else:
        stop = toHVal(stop)

    v = slice(start, stop, step)
    return HSlice.getValueCls()(SLICE, v, 1)



from itertools import product


class TestMatrix():
    """
    Class which instance is a decorator which executes unittest.TestCase
    test method with every combination of argumets
    """

    def __init__(self, *args, **kwargs):
        """
        :note: args, kwargs are lists of arguments which should be passed as a test
            arguments
        """
        self.args = args
        kwargs = sorted(kwargs.items(), key=lambda x: x[0])
        self.kwargs_keys = [x[0] for x in kwargs]
        self.kwargs_values = [x[1] for x in kwargs]
        self.test_arg_values = list(product(*args, *self.kwargs_values))

    def split_args_kwargs(self, args):
        kwargs_cnt = len(self.kwargs_keys)
        if kwargs_cnt:
            _args = args[:kwargs_cnt]
            kwargs = {k: v for k, v in zip(
                self.kwargs_keys, args[:kwargs_cnt])}
            return _args, kwargs
        else:
            return args, {}

    def __call__(self, test_fn):
        test_matrix = self

        def test_wrap(self):
            for args in test_matrix.test_arg_values:
                args, kwargs = test_matrix.split_args_kwargs(args)
                try:
                    test_fn(self, *args, **kwargs)
                except Exception as e:
                    # add note to an exception about which test arguments were
                    # used
                    # import traceback
                    # traceback.print_exc()
                    msg_buff = []
                    for a in args:
                        msg_buff.append(repr(a))

                    for k in test_matrix.kwargs_keys:
                        msg_buff.append("%s=%r" % (k, kwargs[k]))

                    raise Exception(
                        "Test failed %s" % (", ".join(msg_buff)), ) from e

        return test_wrap



from typing import Optional, Tuple

from hwt.serializer.mode import _serializeExclude_eval
from hwt.synthesizer.unit import Unit


class SerializerFilter(object):
    """
    Base class for filters used to exclude some :class:`hwt.synthesizer.unit.Unit` instances from
    target HDL (in order to prevent code duplication, archetype colisions etc.)

    This base implementation keeps track about others objects and calls
    _serializeDecision on the :class:`hwt.synthesizer.unit.Unit` instance to decide if instance should be excluded.
    """
    def __init__(self):
        self.serializedClasses = {}  # type: Type[Unit]: Unit
        # (unitCls, paramsValues) : unitObj
        # where paramsValues are dict name:value
        self.serializedConfiguredUnits = {}

    def do_serialize(self, unit: Unit) -> Tuple[bool, Optional[Unit]]:
        """
        Decide if this unit should be serialized or not eventually fix name
        to fit same already serialized unit

        :param obj: object to serialize
        :param serializedClasses: dict {unitCls : unitobj}
        :param serializedConfiguredUnits: (unitCls, paramsValues) : unitObj
            where paramsValues are named tuple name:value
        """
        assert isinstance(unit, Unit)
        sd = unit._serializeDecision
        if sd is None:
            # the :class:`hwt.synthesizer.unit.Unit` instance does not have any filter function
            return True, None
        else:
            # use UnitInstance filer function
            prevPriv = self.serializedClasses.get(unit.__class__, None)
            do_serialize, nextPriv, replacement = sd(unit, prevPriv)
            self.serializedClasses[unit.__class__] = nextPriv
            return do_serialize, replacement


class SerializerFilterAll(SerializerFilter):
    """
    Ignore any serialization constraints and dump everything
    """
    def do_serialize(self, unit: Unit) -> bool:
        return True, None


class SerializerFilterDoNotExclude(SerializerFilter):
    """
    Use all serialization specifications except @serializeExclude
    Usefull when it is requred to dump all components for sim etc.
    """
    def do_serialize(self, unit: Unit) -> bool:
        orig = unit._serializeDecision
        if orig is _serializeExclude_eval:
            unit._serializeDecision = None
        try:
            return super(SerializerFilterDoNotExclude, self).do_serialize(unit)
        finally:
            unit._serializeDecision = orig



from hwt.doc_markers import internal
from hwt.hdl.hdlObject import HdlObject
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class SignalItem(HdlObject):
    """
    basic hdl signal used to design circuits
    """

    def __init__(self, name, dtype, def_val=None, virtual_only=False):
        """
        :param name: name for better orientation in netlists
            (used only in serialization)
        :param dtype: data type of this signal
        :param def_val: value for initialization
        :param virtual_only: flag indicates that this assignments is only
            virtual and should not be added into
            netlist, because it is only for internal notation
        """
        assert isinstance(name, str), name
        self.name = name
        self._dtype = dtype
        self.virtual_only = virtual_only
        if def_val is None:
            def_val = dtype.from_py(None)
        self.def_val = def_val
        self._set_def_val()

    @internal
    def _set_def_val(self):
        v = self.def_val
        if isinstance(v, RtlSignalBase):
            v = v.staticEval()

        self._val = v.__copy__()



from typing import TypeVar, Generic

from hwt.hdl.types.hdlType import HdlType

T = TypeVar("T", bound=HdlType)


class RtlSignalBase(Generic[T]):
    """
    Main base class for all rtl signals
    """
    pass


class RtlMemoryBase(RtlSignalBase):
    """
    Main base class for all rtl memories
    """
    pass



from hwt.doc_markers import internal
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps, OpDefinition
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BOOL
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import HValue
from pyMathBitPrecise.bit_utils import mask
from pyMathBitPrecise.bits3t import bitsCmp__val, bitsBitOp__val, \
    bitsArithOp__val


# dictionary which hold information how to change operator after
# operands were swapped
CMP_OP_REVERSE = {
    AllOps.EQ: AllOps.EQ,  # (a == b) == (b == a)
    AllOps.NE: AllOps.NE,  # (a != b) == (b != a)
    AllOps.GT: AllOps.LT,  # (a > b)  == (b < a)
    AllOps.LT: AllOps.GT,  # (a < b)  == (b > a)
    AllOps.GE: AllOps.LE,  # (a >= b) == (b <= a)
    AllOps.LE: AllOps.GE,  # (a <= b) == (b >= a)
}


@internal
def bitsCmp_detect_useless_cmp(op0, op1, op):
    v = int(op1)
    width = op1._dtype.bit_length()
    if op0._dtype.signed:
        min_val = -1 if width == 1 else mask(width - 1) - 1
        max_val = 0 if width == 1 else mask(width - 1)
    else:
        min_val = 0
        max_val = mask(width)

    if v == min_val:
        # value can not be lower than min_val
        if op == AllOps.GE:
            # -> always True
            return BOOL.from_py(1, 1)
        elif op == AllOps.LT:
            # -> always False
            return BOOL.from_py(0, 1)
        elif op == AllOps.LE:
            # convert <= to == to highlight the real function
            return AllOps.EQ
    elif v == max_val:
        # value can not be greater than max_val
        if op == AllOps.GT:
            # always False
            return BOOL.from_py(0, 1)
        elif op == AllOps.LE:
            # always True
            return BOOL.from_py(1, 1)
        elif op == AllOps.GE:
            # because value can not be greater than max
            return AllOps.EQ


@internal
def bitsCmp(self, other, op, selfReduceVal, evalFn=None):
    """
    Apply a generic comparison binary operator

    :attention: If other is Bool signal convert this to bool (not ideal,
        due VHDL event operator)
    :ivar self: operand 0
    :ivar other: operand 1
    :ivar op: operator used
    :ivar selfReduceVal: the value which is a result if operands are all same signal (e.g. a==a = 1, b<b=0)
    :ivar evalFn: override of a python operator function (by default one from "op" is used)
    """
    t = self._dtype
    other = toHVal(other, t)
    ot = other._dtype
    if not isinstance(ot, t.__class__):
        raise TypeError(ot)

    if evalFn is None:
        evalFn = op._evalFn

    iamVal = isinstance(self, HValue)
    otherIsVal = isinstance(other, HValue)
    type_compatible = False
    if ot == BOOL:
        self = self._auto_cast(BOOL)
        type_compatible = True
    elif t == ot:
        type_compatible = True
    # lock type widht/signed to other type with
    elif not ot.strict_width or not ot.strict_sign:
        type_compatible = True
        other = other._auto_cast(t)
    elif not t.strict_width or not t.strict_sign:
        type_compatible = True
        other = other._auto_cast(ot)

    if iamVal and otherIsVal:
        if type_compatible:
            return bitsCmp__val(self, other, evalFn)
    else:
        if type_compatible:
            # try to reduce useless cmp
            res = None
            if otherIsVal and other._is_full_valid():
                res = bitsCmp_detect_useless_cmp(self, other, op)
            elif iamVal and self._is_full_valid():
                res = bitsCmp_detect_useless_cmp(other, self, CMP_OP_REVERSE[op])

            if res is None:
                pass
            elif isinstance(res, HValue):
                return res
            else:
                assert res == AllOps.EQ, res
                op = res

            if self is other:
                return selfReduceVal
            else:
                return Operator.withRes(op, [self, other], BOOL)

        elif t.signed != ot.signed:
            if t.signed is None:
                self = self._convSign(ot.signed)
                return bitsCmp(self, other, op, evalFn)
            elif ot.signed is None:
                other = other._convSign(t.signed)
                return bitsCmp(self, other, op, evalFn)

    raise TypeError(f"Values of types ({self._dtype}, {other._dtype}) are not comparable")

@internal
def bitsBitOp(self, other, op: OpDefinition, getVldFn, reduceCheckFn, selfReduceVal):
    """
    Apply a generic bitwise binary operator

    :attention: If other is Bool signal, convert this to bool
        (not ideal, due VHDL event operator)
    :ivar self: operand 0
    :ivar other: operand 1
    :ivar op: operator used
    :ivar getVldFn: function to resolve invalid (X) states
    :ivar reduceCheckFn: function to reduce useless operators (partially evaluate the expression if possible)
    :ivar selfReduceVal: the value which is a result if operands are all same signal (e.g. a&a = a, b^b=0)
    """
    other = toHVal(other, self._dtype)

    iamVal = isinstance(self, HValue)
    otherIsVal = isinstance(other, HValue)

    if iamVal and otherIsVal:
        other = other._auto_cast(self._dtype)
        return bitsBitOp__val(self, other, op._evalFn, getVldFn)
    else:
        s_t = self._dtype
        o_t = other._dtype
        if not isinstance(o_t, s_t.__class__):
            raise TypeError(o_t)

        if s_t == o_t:
            pass
        elif o_t == BOOL and s_t != BOOL:
            self = self._auto_cast(BOOL)
            return op._evalFn(self, other)
        elif o_t != BOOL and s_t == BOOL:
            other = other._auto_cast(BOOL)
            return op._evalFn(self, other)
        else:
            if s_t.signed is not o_t.signed and bool(s_t.signed) == bool(o_t.signed):
                # automatically cast unsigned to vector
                if s_t.signed == False and o_t.signed is None:
                    self = self._vec()
                    s_t = self._dtype
                elif s_t.signed is None and o_t.signed == False:
                    other = other._vec()
                    o_t = other._dtype
                else:
                    raise ValueError("Invalid value for signed flag of type", s_t.signed, o_t.signed, s_t, o_t)

            if s_t == o_t:
                # due to previsous cast the type may become the same
                pass
            elif s_t.bit_length() == 1 and o_t.bit_length() == 1\
                    and s_t.signed is o_t.signed \
                    and s_t.force_vector != o_t.force_vector:
                # automatically cast to vector with a single item to a single bit
                if s_t.force_vector:
                    self = self[0]
                else:
                    other = other[0]

            else:
                raise TypeError("Can not apply operator %r (%r, %r)" %
                                (op, self._dtype, other._dtype))

        if otherIsVal:
            r = reduceCheckFn(self, other)
            if r is not None:
                return r

        elif iamVal:
            r = reduceCheckFn(other, self)
            if r is not None:
                return r

        elif self is other:
            return selfReduceVal

        return Operator.withRes(op, [self, other], self._dtype)


@internal
def bitsArithOp(self, other, op):
    other = toHVal(other, self._dtype)
    if not isinstance(other._dtype, Bits):
        raise TypeError(other._dtype)

    self_is_val = isinstance(self, HValue)
    other_is_val = isinstance(other, HValue)

    if self_is_val and other_is_val:
        return bitsArithOp__val(self, other, op._evalFn)
    else:
        if self._dtype.signed is None:
            self = self._unsigned()

        if other_is_val and other._is_full_valid() and int(other) == 0:
            return self

        resT = self._dtype
        if self_is_val and self._is_full_valid() and int(self) == 0:
            return other._auto_cast(resT)

        if isinstance(other._dtype, Bits):
            t0 = self._dtype
            t1 = other._dtype
            if t0.bit_length() != t1.bit_length():
                if not t1.strict_width:
                    # resize to type of this
                    other = other._auto_cast(t1)
                    t1 = other._dtype
                    pass
                elif not t0.strict_width:
                    # resize self to type of result
                    self = self._auto_cast(t0)
                    t0 = self._dtype
                    pass
                else:
                    raise TypeError("%r %r %r" % (self, op, other))

            if t1.signed != resT.signed:
                other = other._convSign(t0.signed)
        else:
            raise TypeError("%r %r %r" % (self, op, other))

        o = Operator.withRes(op, [self, other], self._dtype)
        return o._auto_cast(resT)



from copy import copy
from decimal import DecimalTuple
import math

from hwt.doc_markers import internal
from hwt.hdl.value import HValue
from pyMathBitPrecise.floatt import FloattVal


class HFloatVal(HValue, FloattVal):
    """
    HValue class for HFloat type
    """

    @classmethod
    def from_py(cls, typeObj, val, vld_mask=None):
        assert vld_mask is None, vld_mask
        if isinstance(val, int):
            val = float(val)
            if float(val) != val:
                raise NotImplementedError("Need to implement better conversion method")

        if val is None:
            sign = 0
            exp = 0
            man = 0
            assert vld_mask is None or vld_mask == 0
            vld_mask = 0
        elif isinstance(val, float):
            man, exp = math.frexp(val)
            man = abs(man)
            man = int(man * (2 ** typeObj.mantisa_w))
            sign = int(val < 0)
            if vld_mask is None:
                vld_mask = 1
        elif isinstance(val, tuple):
            sign, man, exp = val
            if vld_mask is None:
                vld_mask = 1
        else:
            raise TypeError(val)

        return cls(typeObj, DecimalTuple(sign, man, exp), vld_mask)

    def _is_full_valid(self):
        return self.vld_mask == 1

    def to_py(self):
        """
        Convert to python slice object
        """
        return float(self)

    def _eq_val(self, other):
        assert isinstance(other, HFloatVal)
        return self.val == other.val

    def _eq(self, other):
        return self._eq__val(other)

    def __copy__(self):
        v = HValue.__copy__(self)
        v.val = copy(v.val)
        return v

    @internal
    def __hash__(self):
        v = self.val
        return hash((self._dtype, v))



from types import MethodType
from typing import Tuple, Set, Optional

from hwt.doc_markers import internal
from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.hObjList import HObjList
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase, InterfaceBase
from hwt.synthesizer.param import Param
from hdlConvertorAst.translate.common.name_scope import WithNameScope


@internal
def nameAvailabilityCheck(obj, propName, prop):
    """
    Check if not redefining property on obj
    """
    if getattr(obj, propName, None) is not None:
        p = getattr(obj, propName)
        raise IntfLvlConfErr(f"{obj} already has property {propName:s} old:{p} new:{prop}")


@internal
class MakeParamsShared(object):
    """
    All newly added interfaces and units will share all parametes with unit
    specified in constructor of this object.
    """

    def __init__(self, unit, exclude, prefix):
        self.unit = unit
        self.exclude = exclude
        self.prefix = prefix

    def __enter__(self):
        orig = self.unit._setAttrListener
        self.orig = orig
        exclude = self.exclude
        prefix = self.prefix

        def MakeParamsSharedWrap(self, iName, i):
            if isinstance(i, (InterfaceBase, UnitBase, HObjList)):
                i._updateParamsFrom(self, exclude=exclude, prefix=prefix)
            return orig(iName, i)

        self.unit._setAttrListener = MethodType(MakeParamsSharedWrap,
                                                self.unit)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.unit._setAttrListener = self.orig


@internal
class MakeClkRstAssociations(object):
    """
    All newly added interfaces will be associated with clk, rst
    specified in constructor of this object.
    """

    def __init__(self, unit, clk=None, rst=None):
        self.unit = unit
        self.clk = clk
        self.rst = rst

    def __enter__(self):
        orig = self.unit._setAttrListener
        self.orig = orig
        clk = self.clk
        rst = self.rst

        def MakeClkRstAssociationsWrap(self, iName, i):
            if isinstance(i, (InterfaceBase, HObjList)):
                i._make_association(clk=clk, rst=rst)
            return orig(iName, i)

        self.unit._setAttrListener = MethodType(MakeClkRstAssociationsWrap,
                                                self.unit)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.unit._setAttrListener = self.orig


class PropDeclrCollector(object):
    """
    Class which manages the registration of components and interfaces
    in specified elaboration phases.

    It uses __setattr__ listeners to detect new properties and then calls
    a litener function to process the registration.

    Used for Unit, Interface classes to detect and load interfaces and components.
    """

    def _config(self) -> None:
        """
        Configure object parameters

        * setup all parameters on this object,
          use Param class instances to allow use of parameter inheritance
        * called in __init__ of class
        """
        pass

    def _declr(self) -> None:
        """
        In this function user should specify the declaration of interfaces for communication with outside word.
        It is also better to declare sub components there as it allows for better paralelization during the build.

        * _declr method is called after _config
        * if this object is :class:`hwt.synthesizer.unit.Unit` all interfaces are threated as externally accessible interfaces
          if this object is Interface instance all subinterfaces are loaded as well
        """
        pass

    def _impl(self) -> None:
        """
        Implementation - construct main body of the component in this function.

        * called after _declr
        """
        pass

    @internal
    def __setattr__(self, attr, value) -> None:
        """setattr with listener injector"""
        try:
            saListerner = self._setAttrListener
        except AttributeError:
            super().__setattr__(attr, value)
            return

        if saListerner:
            value = saListerner(attr, value)
        super().__setattr__(attr, value)

    # configuration phase
    @internal
    def _loadConfig(self) -> None:
        """
        Load params in _config()
        """
        if not hasattr(self, '_params'):
            self._params = []

        self._setAttrListener = self._paramCollector
        self._config()
        self._setAttrListener = None

    @internal
    def _registerParameter(self, pName, parameter: Param) -> None:
        """
        Register Param object on interface level object
        """
        nameAvailabilityCheck(self, pName, parameter)
        # resolve name in this scope
        assert parameter._name is None, (
            "Param object is already assigned to %r.%s"
            % (parameter.unit, parameter._name))
        # add name in this scope
        parameter._name = pName
        parameter._parent = self

        self._params.append(parameter)

    def _paramsShared(self,
                      exclude: Optional[Tuple[Set[str], Set[str]]]=None,
                      prefix="") -> MakeParamsShared:
        """
        Auto-propagate params by name to child components and interfaces
        Usage:

        .. code-block:: python

            with self._paramsShared():
                # your interfaces and unit which should share all params with "self" there

        :param exclude: tuple (src param names to exclude, dst param names to exclude)
        :param prefix: prefix which should be added to name of child parameters
            before parameter name matching
        """
        return MakeParamsShared(self, exclude=exclude, prefix=prefix)

    def _make_association(self, clk=None, rst=None) -> None:
        """
        Associate this object with specified clk/rst
        """
        if clk is not None:
            assert self._associatedClk is None, ("Already associated with clock", self._associatedClk)
            self._associatedClk = clk

        if rst is not None:
            assert self._associatedRst is None, ("Already associated with reset", self._associatedRst)
            self._associatedRst = rst

    def _associated(self, clk=None, rst=None) -> MakeClkRstAssociations:
        """
        associate newly added interfaces to "self" with selected clk, rst
        (if interface is not associated agents try to find clk/rst by _getAssociatedClk/_getAssociatedRst
        which will search for any clk/rst on parent recursively)
        Usage:

        .. code-block:: python

            with self._associated(clk=self.myClk, rst=self.myRst):
                self.myAxi = AxiStrem()
                # this interface is associated with myClk and myRst
                # simulation agents and component builders will use them


        :param exclude: params which should not be shared
        """
        return MakeClkRstAssociations(self, clk, rst)

    def _updateParamsFrom(self,
                          otherObj: "PropDeclrCollector",
                          updater,
                          exclude: Optional[Tuple[Set[str], Set[str]]],
                          prefix: str) -> "PropDeclrCollector":
        """
        Update all parameters which are defined on self from otherObj

        :param otherObj: other object which Param instances should be updated
        :param updater: updater function(self, myParameter,
                                         onOtherParameterName, otherParameter)
        :param exclude: tuple of set of param names for src and dst which
                        which should be excluded
        :param prefix: prefix which should be added to name of paramters
                       of this object before matching parameter name on parent
        """
        excluded_src = set()
        excluded_dst = set()
        if exclude is not None:
            exclude_src = set(exclude[0])
            exclude_dst = set(exclude[1])

        for myP in self._params:
            if exclude is not None and myP._name in exclude_dst:
                excluded_dst.add(myP._name)
                continue
            pPName = prefix + myP._name
            try:
                otherP = getattr(otherObj, pPName)
                # if not isinstance(otherP, Param):
                #     continue
            except AttributeError:
                continue

            if exclude is not None and pPName in exclude_src:
                excluded_src.add(pPName)
                continue
            updater(self, myP, otherP)

        if exclude is not None:
            # assert that what should be excluded really exists
            assert exclude_src == excluded_src, (exclude_src, excluded_src)
            assert exclude_dst == excluded_dst, (exclude_dst == excluded_dst)
        return self

    # declaration phase
    @internal
    def _registerUnit(self, uName, unit):
        """
        Register unit object on interface level object
        """
        nameAvailabilityCheck(self, uName, unit)
        assert unit._parent is None
        unit._parent = self
        unit._name = uName
        self._units.append(unit)

    @internal
    def _registerInterface(self, iName, intf, isPrivate=False):
        """
        Register interface object on interface level object
        """
        nameAvailabilityCheck(self, iName, intf)
        assert intf._parent is None
        intf._parent = self
        intf._name = iName
        intf._ctx = self._ctx

        # _setAsExtern() not used because _interfaces are not intitialized yet
        if isPrivate:
            self._private_interfaces.append(intf)
            intf._isExtern = False
        else:
            self._interfaces.append(intf)
            intf._isExtern = True

    @internal
    def _declrCollector(self, name, prop):
        if name in ("_associatedClk", "_associatedRst"):
            object.__setattr__(self, name, prop)
            return prop

        if isinstance(prop, InterfaceBase):
            self._registerInterface(name, prop)
        elif isinstance(prop, UnitBase):
            self._registerUnit(name, prop)
        elif isinstance(prop, HObjList):
            self._registerArray(name, prop)
        return prop

    @internal
    def _registerArray(self, name, items: HObjList):
        """
        Register array of items on interface level object
        """
        items._parent = self
        items._name = name
        items._on_append = self._registerArray_append
        for i, item in enumerate(items):
            self._registerArray_append(items, item, i)

    @internal
    def _registerArray_append(self, h_obj_list: HObjList, item, index: int):
        """
        Register a single object in the list
        """
        setattr(self, "%s_%d" % (h_obj_list._name, index), item)

    # implementation phase
    @internal
    def _loadImpl(self):
        self._setAttrListener = self._implCollector
        self._impl()
        self._setAttrListener = None

    @internal
    def _registerUnitInImpl(self, uName, u):
        """
        :attention: unit has to be parametrized before it is registered
            (some components can change interface by parametrization)
        """
        self._registerUnit(uName, u)
        u._loadDeclarations()
        sm = self._store_manager
        with WithNameScope(sm, sm.name_scope.parent):
            self._lazy_loaded.extend(u._to_rtl(
                self._target_platform, self._store_manager))
        u._signalsForSubUnitEntity(self._ctx, "sig_" + uName)

    @internal
    def _registerIntfInImpl(self, iName, i):
        """
        Register interface in implementation phase
        """
        raise NotImplementedError()

    @internal
    def _paramCollector(self, pName, prop):
        if isinstance(prop, Param):
            self._registerParameter(pName, prop)
            return prop._initval
        else:
            return prop

    @internal
    def _implCollector(self, name, prop):
        """
        Handle property definitions in _impl phase
        """
        if isinstance(prop, InterfaceBase):
            if prop._parent is self:
                return prop
            self._registerIntfInImpl(name, prop)
        elif isinstance(prop, UnitBase):
            if prop._parent is self:
                return prop
            self._registerUnitInImpl(name, prop)
        elif isinstance(prop, HObjList):
            if prop._parent is self:
                return prop
            self._registerArray(name, prop)
        return prop




import sys
from typing import Union

from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.value import HValue
from hwt.simulator.rtlSimulator import BasicRtlSimulatorWithSignalRegisterMethods
from hwtSimApi.basic_hdl_simulator.proxy import BasicRtlSimProxy
from hwtSimApi.basic_hdl_simulator.sim_utils import ValueUpdater, \
    ArrayValueUpdater
from pyDigitalWaveTools.vcd.writer import VcdWriter
#from pyMathBitPrecise.array3t import Array3t
from pyMathBitPrecise.bits3t import Bits3t
from pyMathBitPrecise.enum3t import Enum3t


class BasicRtlSimulatorVcd(BasicRtlSimulatorWithSignalRegisterMethods):
    supported_type_classes = (Bits, HEnum, Bits3t, Enum3t,
                              #Array3t
                              )

    def create_wave_writer(self, file_name):
        self.wave_writer = VcdWriter(open(file_name, "w"))
        self.logChange = self._logChange

    def finalize(self):
        # because set_trace_file() may not be called
        # and it this case the vcd config is not set
        if self.wave_writer is None:
            return

        f = self.wave_writer._oFile
        if f not in (sys.__stderr__, sys.__stdin__, sys.__stdout__):
            f.close()

    def _logChange(self, nowTime: int,
                  sig: BasicRtlSimProxy,
                  nextVal: HValue,
                  valueUpdater: Union[ValueUpdater, ArrayValueUpdater]):
        """
        This method is called for every value change of any signal.
        """
        try:
            self.wave_writer.logChange(nowTime, sig, nextVal, valueUpdater)
        except KeyError:
            # not every signal has to be registered
            # (if it is not registered it means it is ignored)
            pass



from typing import Union

from hdlConvertorAst.hdlAst._expr import HdlValueId, HdlOp, HdlOpType
from hdlConvertorAst.to.hdlUtils import bit_string
from hdlConvertorAst.to.verilog.constants import SIGNAL_TYPE
from hdlConvertorAst.translate.verilog_to_basic_hdl_sim_model.utils import hdl_call, \
    hdl_getattr
from hdlConvertorAst.translate.common.name_scope import LanguageKeyword
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.statements.assignmentContainer import HdlAssignmentContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import SLICE
from hwt.hdl.types.enumVal import HEnumVal
from hwt.hdl.value import HValue
from hwt.hdl.variables import SignalItem
from hwt.serializer.generic.value import ToHdlAst_Value
from hwt.serializer.hwt.ops import ToHdlAstHwt_ops
from hwt.serializer.systemC.utils import systemCTypeOfSig
from hwt.serializer.verilog.context import SignalTypeSwap
from hwt.serializer.verilog.ops import ToHdlAstVerilog_ops
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class ToHdlAstSystemC_expr(ToHdlAst_Value):
    static_cast = HdlValueId("static_cast", obj=LanguageKeyword())
    op_transl_dict = ToHdlAstVerilog_ops.op_transl_dict

    def as_hdl_Value(self, v):
        if isinstance(v, tuple):
            return tuple((self.as_hdl(o2) for o2 in v))
        return super(ToHdlAstSystemC_expr, self).as_hdl_Value(v)

    def as_hdl_operand(self, operand: Union[RtlSignal, HValue], i: int,
                       operator: Operator):
        return self.as_hdl(operand)

    def as_hdl_Operator(self, op: Operator):
        ops = op.operands
        o = op.operator

        if o == AllOps.INDEX:
            assert len(ops) == 2
            o0, o1 = ops
            if o1._dtype == SLICE:
                # index to .range(x, y)
                o0_hdl = self.as_hdl_operand(o0, 0, op)
                o0_hdl = hdl_getattr(o0_hdl, "range")
                return hdl_call(o0_hdl, [self.as_hdl_Value(o1.val.start),
                                         self.as_hdl_Value(o1.val.stop)])
            else:
                return ToHdlAstVerilog_ops.as_hdl_Operator(self, op)
        elif o in ToHdlAstHwt_ops._cast_ops:
            assert len(ops) == 1, ops
            t = self.as_hdl_HdlType(op.result._dtype)
            return hdl_call(
                HdlOp(HdlOpType.PARAMETRIZATION, [self.static_cast, t]),
                [self.as_hdl_Value(ops[0]), ])
        elif o == AllOps.CONCAT:
            isNew, o = self.tmpVars.create_var_cached("tmpConcat_",
                                                      op.result._dtype,
                                                      postponed_init=True,
                                                      extra_args=(AllOps.CONCAT, op.result))
            if isNew:
                o.drivers.append(HdlAssignmentContainer(op, o, virtual_only=True))
                self.tmpVars.finish_var_init(o)

            return self.as_hdl(o)
        else:
            return ToHdlAstVerilog_ops.as_hdl_Operator(self, op)

    def as_hdl_SignalItem(self, si: SignalItem, declaration=False):
        if declaration:
            sigType = systemCTypeOfSig(si)
            with SignalTypeSwap(self, sigType):
                return ToHdlAst_Value.as_hdl_SignalItem(self, si, declaration=True)
        else:
            if si.hidden and si.origin is not None:
                return self.as_hdl(si.origin)
            else:
                sigType = systemCTypeOfSig(si)
                _si = HdlValueId(si.name, obj=si)
                if self._in_sensitivity_list or self._is_target or sigType is SIGNAL_TYPE.REG:
                    return _si
                else:
                    return hdl_call(hdl_getattr(_si, "read"), [])

    def as_hdl_BitsVal(self, val):
        t = val._dtype
        w = t.bit_length()
        _v = bit_string(val.val, w, val.vld_mask)
        t = self.as_hdl_HdlType_bits(Bits(w, signed=t.signed))
        return hdl_call(t, [_v, ])

    def as_hdl_HEnumVal(self, val: HEnumVal):
        i = val._dtype._allValues.index(val.val)
        assert i >= 0
        return self.as_hdl_int(i)

    def as_hdl_HArrayVal(self, val):
        return [self.as_hdl_Value(v) for v in val]
